import os
import glob
import errno
import signal
import json
import sys
import inspect
import shutil
import subprocess
import time
import re
import multiprocessing as mp
import pprint

from multiprocessing import Process
from concurrent.futures import ThreadPoolExecutor
from typing import Text, List, Dict, Union
from importlib import import_module
from datetime import date, datetime
from collections import OrderedDict

import grpc
import jsbeautifier

from google.protobuf.message import Message
from appyratus.utils.string_utils import StringUtils
from appyratus.utils.func_utils import FuncUtils
from appyratus.utils.dict_utils import DictUtils, DictObject
from appyratus.utils.time_utils import TimeUtils
from appyratus.utils.template_utils import TemplateEnvironment

from ravel.util import (
    is_resource,
    is_batch,
    get_class_name,
    is_port_in_use,
)
from ravel.util.loggers import console
from ravel.util.json_schema import JsonSchemaGenerator
from ravel.app.base import Application
from ravel.schema import Schema, fields

from .grpc_method import GrpcMethod
from .grpc_client import GrpcClient
from .util import get_stripped_schema_name
from .proto import MessageGenerator
from .constants import (
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_GRACE,
    DEFAULT_IS_SECURE,
    REGISTRY_PROTO_FILE,
    PB2_MODULE_PATH_FSTR,
    PB2_GRPC_MODULE_PATH_FSTR,
    PROTOC_COMMAND_FSTR,
    PROTOC_WEB_GENERATE_CLIENT_COMMAND,
    PROTOC_WEB_GENERATE_MESSAGE_CLASSES_COMMAND,
)


class GrpcOptionsSchema(Schema):
    """
    Manifest gRPC options schema
    """
    client_host = fields.String(required=True, default=DEFAULT_HOST)
    server_host = fields.String(required=True, default=DEFAULT_HOST)
    secure_channel = fields.Bool(required=True, default=DEFAULT_IS_SECURE)
    port = fields.String(required=True, default=DEFAULT_PORT)
    grace = fields.Float(required=True, default=DEFAULT_GRACE)


class GrpcService(Application):
    """
    Grpc server.

    All configuration options and data structures pertaining to gRPC itself are
    kept in self.grpc. For the most part, this object is constructed during
    bootstrap.

    It is possible to generate a gRPC client for the service via the "client"
    property. The service must be bootstrapped before accessing the client.

    Note that gRPC manages its server as a thread pool. When we call
    GrpcService.start, we are really just starting the threads in the pool and
    then putting the main thread in a sleep lock.
    """

    def __init__(self, middleware=None, *args, **kwargs):
        super().__init__(middleware=middleware, *args, **kwargs)
        self.grpc = DictObject()
        self.grpc.options = DictObject()
        self.grpc.worker_process_count = mp.cpu_count()
        self.grpc.worker_thread_pool_size = 1
        self.grpc.pb2 = None
        self.grpc.pb2_grpc = None
        self.grpc.server = None
        self.grpc.servicer = None
        self._client = None

    @property
    def client(self) -> GrpcClient:
        """
        Return a memoized GrpcClient instance. However, if the service isn't
        bootstrapped, None will be returned, as the client derives itself
        from a bootstrapped service.
        """
        if not self.is_bootstrapped:
            return None
        elif self._client is None:
            self._client = GrpcClient(self)
        return self._client

    @property
    def action_type(self):
        return GrpcMethod

    def on_bootstrap(self, options: Dict = None, build=False):
        """
        Compute gRPC configuration variables, build the proto file and grpcio
        "pb2" Python modules, and compute other metadata needed to route
        requests to RPC API methods.
        """
        self._on_bootstrap_configure_grpc(options)
        self._on_bootstrap_build_and_import_grpc_modules(build)
        self._on_bootstrap_aggregate_response_message_types()

    def build(self, manifest=None, options: Dict = None):
        self.bootstrap(manifest=manifest, build=True, options=options)

    def build_js(self, manifest: Text, dest: Text):
        """
        Generate JavaScript files containing the porotobuf message types and
        client stub.
        """
        # TODO: move these methods into a grpc_web module
        self.bootstrap(manifest=manifest, build=False)

        # ensure template build directory exists
        dest = os.path.join(dest, 'grpc')
        try:
            os.makedirs(dest)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        self._compile_grpc_web_message_classes(dest)
        self._compile_grpc_web_client_stub(dest)
        self._render_grpc_web_client_files(dest)

    def _render_grpc_web_client_files(self, dest):
        dest = os.path.realpath(dest)

        # generate JSON schemas for all ravel schemas
        # referenced in this app
        def unnest(schema, unnested):
            for child_schema in schema.children:
                child_schema_name = get_stripped_schema_name(child_schema)
                if child_schema_name in unnested:
                    console.warning(
                        message=(
                            f'gRPC web client build process '
                            f'ignoring schema with duplicate name'
                        ),
                        data={
                            'schema': child_schema_name,
                            'parent_schema': get_stripped_schema_name(schema),
                        }
                    )
                unnested[child_schema_name] = schema
                unnest(child_schema, unnested)

        nested_schemas = OrderedDict()
        for resource_type in self.res.values():
            unnest(resource_type.ravel.schema, nested_schemas)

        action_schemas = OrderedDict()
        for action in self.actions.values():
            for schema in action.schemas.values():
                action_schemas[get_stripped_schema_name(schema)] = schema

        # initialize templating engine
        template_dir = os.path.join(os.path.dirname(__file__), 'templates')
        jinja = TemplateEnvironment(template_dir)
        template_context = {
            'list': list,
            'dict': dict,
            'str': str,
            'int': int,
            'set': set,
            'len': len,
            'app': self,
            'inspect': inspect,
            'nested_schemas': nested_schemas,
            'action_schemas': action_schemas,
            'json_schema_generator': JsonSchemaGenerator(),
            'class_name': get_class_name,
        }

        console.info(
            message=f'jinja2 template directory',
            data={'path': template_dir}
        )

        # render templates
        for js_fname in glob.glob(template_dir + '/*.js'):
            js_fname = os.path.basename(js_fname)
            with open(os.path.join(dest, js_fname), 'w') as js_file:
                # render template from file
                template = jinja.from_filename(js_fname)
                js_source = template.render(template_context)
                # lint the generated JS code
                formated_js_source = jsbeautifier.beautify(js_source)
                # write to dest directory
                console.info(f'rendering {dest}/{os.path.basename(js_fname)}')
                js_file.write(formated_js_source)

    def on_decorate(self, action):
        pass

    def on_request(self, action, request, *args, **kwargs):
        """
        Take the attributes on the incoming protobuf Message object and
        map them to the args and kwargs expected by the action target.
        """
        console.debug(
            message=f'RPC method {action.name} requested',
        )

        # get field data to process into args and kwargs
        def deserialize_value(field, value):
            if isinstance(field, fields.Dict):
                if not value:
                    value = None if field.nullable else {}
                else:
                    value = self.json.decode(value)
            return value

        all_arguments = {}
        for field in action.schemas.request.fields.values():
            value = getattr(request, field.name, None)
            all_arguments[field.name] = deserialize_value(field, value)

        # add a dummy request object so that partition_arguments can
        # run below; then remove it from the resulting args tuple.
        all_arguments['request'] = None

        # process field_data into args and kwargs
        args, kwargs = FuncUtils.partition_arguments(
            action.signature, all_arguments
        )

        args = args[1:]

        return (args, kwargs)

    def on_response(self, action, result, *raw_args, **raw_kwargs):
        """
        Map the return dict from the action to the expected outgoing protobuf
        response Message object.
        """
        response_type = self.grpc.response_types[action.name]
        response = response_type()
        response_data = None

        def validate(data):
            data, errors = schema.process(data)
            if errors:
                console.error(
                    message=f'response validation errors',
                    data={
                        'data': pprint.pformat(data),
                        'errors': errors,
                    }
                )
            return data

        if result:
            schema = action.schemas.response
            dumped_result = self._dump_result_obj(result)
            if isinstance(dumped_result, list):
                response_data = [validate(x) for x in dumped_result]
            else:
                response_data = validate(dumped_result)

        if response_data:
            if isinstance(response_data, list):
                response = iter(
                    self.build_response_message(response_type(), data)
                    for data in response_data
                )
            else:
                response = self.build_response_message(response, response_data)

        return response

    @staticmethod
    def on_start_worker(app):
        """
        This method runs as the "initializer" for each thread in the grpc
        server's thread pool. Override in subclass.
        """

    def on_start(self):
        """
        Start the RPC client or server.
        """
        # try to ensure the desired port is available
        if is_port_in_use(self.grpc.options.server_address):
            console.critical(
                f'gRPC service address {self.grpc.options.server_address} '
                f'in use. trying to kill stale service processes.'
            )
            server_port = self.grpc.options.port
            try:
                kill_cmd = f'lsof -ti tcp:{server_port} | xargs kill'
                console.info(
                    message=f'running command to free gRPC service port',
                    data={'port': server_port, 'command': kill_cmd}
                )
                subprocess.getoutput(kill_cmd)
            except Exception:
                console.exception(
                    f'error running command to free gRPC service port'
                )
                exit(-1)
            if is_port_in_use(self.grpc.options.server_address):
                console.critical('could not start gRPC service')
                exit(-1)

        console.info(
            message='strating gRPC server. Press ctrl+c to stop.',
            data={
                'address': self.grpc.options.server_address,
                'methods': list(self.actions.keys()),
                'worker_process_count': self.grpc.worker_process_count,
                'worker_thread_pool_size': self.grpc.worker_thread_pool_size,
            }
        )
        # start the server process/es
        if self.grpc.worker_process_count == 1:
            # serve in main forground process
            self._init_server_and_run()
        else:
            # serve in forked subprocesses
            self._spawn_worker_processes()

    def _spawn_worker_processes(self):
        processes = []

        # entrypoint or "target" function for server processes
        def entrypoint(service):
            service.bootstrap(force=True)
            service._init_server_and_run()

        # create and run the server processes
        for _ in range(self.grpc.worker_process_count):
            process = Process(target=entrypoint, args=(self, ))
            processes.append(process)
            process.start()

        self._setup_os_signal_handling_when_multiprocessing()

        for process in processes:
            process.join(timeout=5)

    def _setup_os_signal_handling_when_multiprocessing(self):
        def handler(signum, frame):
            console.debug(
                'ignoring SIGINT in root process. joining '
                'all server subprocesses.'
            )
        # set the signal handler
        signal.signal(signal.SIGINT, handler)

    def _init_server_and_run(self):
        # the grpc server runs in a thread pool
        self.grpc.executor = ThreadPoolExecutor(
            max_workers=self.grpc.worker_thread_pool_size,
            initializer=self.on_start_worker,
            initargs=(self, )
        )

        for i in range(self.grpc.worker_thread_pool_size):
            self.grpc.executor.submit(lambda: None)

        # build grpc server
        self.grpc.server = grpc.server(self.grpc.executor)
        self.grpc.server.add_insecure_port(self.grpc.options.server_address)

        # build grpcio "servicer" class
        self.grpc.servicer = self._grpc_servicer_factory()
        self.grpc.servicer.add_to_grpc_server(
            self.grpc.servicer, self.grpc.server
        )

        self.grpc.server.start()

        # sleep-lock the main thread to keep the server running,
        # as the server runs as a daemon.
        try:
            while True:
                time.sleep(9999)
        except KeyboardInterrupt:
            # stop server on ctrl+c
            if self.grpc.server is not None:
                console.info(f'stopping gRPC server (PID: {os.getpid()})')
                self.grpc.server.stop(grace=self.grpc.options.grace)

    def _on_bootstrap_configure_grpc(self, options: Dict):
        """
        Compute the gRPC options dict and other values needed to run this
        application as a gRPC service.
        """
        grpc = self.grpc

        # get the python package containing this Application
        pkg_path = self.manifest.package
        pkg = import_module(self.manifest.package)

        # compute some file and module paths
        grpc.pkg_dir = os.path.dirname(pkg.__file__)
        grpc.build_dir = os.path.join(grpc.pkg_dir, 'grpc')
        grpc.proto_file = os.path.join(grpc.build_dir, REGISTRY_PROTO_FILE)

        # compute grpc options dict from options kwarg and manifest data
        kwarg_options = options or {}
        manifest_options = self.manifest.data.get('grpc', {})
        computed_options = DictUtils.merge(kwarg_options, manifest_options)

        # generate default values into gRPC options data
        schema = GrpcOptionsSchema()
        options, errors = schema.process(computed_options)
        if errors:
            raise Exception(f'invalid gRPC options: {errors}')

        grpc.options = DictObject(options)
        grpc.options.port = port = options['port']
        grpc.options.server_host = server_host = options['server_host']
        grpc.options.client_host = client_host = options['client_host']
        grpc.options.server_address = (f'{server_host}:{port}')
        grpc.options.client_address = (f'{client_host}:{port}')
        grpc.pb2_mod_path = PB2_MODULE_PATH_FSTR.format(pkg_path)
        grpc.pb2_grpc_mod_path = PB2_GRPC_MODULE_PATH_FSTR.format(pkg_path)

        # add build directory to PYTHONPATH; otherwise, gRPC won't know where
        # the pb2 modules are when it tries to import them.
        sys.path.append(grpc.build_dir)

    def _on_bootstrap_build_and_import_grpc_modules(self, build: bool):
        """
        Call an external grpcio command to generate the gRPC python modules
        necessary for running this service. Then we import them, because we
        need to derive client and servicer subclasses and procedures from the
        autogenerated base components contained therein.
        """
        grpc = self.grpc
        pkg_path = self.manifest.package

        # try to import the pb2 module just to see if it exists
        try:
            import_module(grpc.pb2_mod_path, pkg_path)
            pb2_module_dne = False    # dne: does not exist
        except Exception:
            console.warning(f'{grpc.pb2_mod_path} does not exist. rebuilding')
            pb2_module_dne = True

        # build the pb2 and pb2_grpc modules
        if build or pb2_module_dne:
            self._build_pb2_modules()
            time.sleep(0.25)

        # now import the dynamically-generated pb2 modules
        grpc.pb2 = import_module(grpc.pb2_mod_path, pkg_path)
        grpc.pb2_grpc = import_module(grpc.pb2_grpc_mod_path, pkg_path)

    def _on_bootstrap_aggregate_response_message_types(self):
        """
        Build a lookup table of protobuf response Message types for use when
        routing requests to their downstream actions.
        """
        grpc = self.grpc
        grpc.response_types = {}
        for action in self.actions.values():
            schema_type_name = get_class_name(action.schemas.response)
            message_type_name = get_stripped_schema_name(
                StringUtils.camel(schema_type_name)
            )
            response_message_type = getattr(self.grpc.pb2, message_type_name)
            grpc.response_types[action.name] = response_message_type

    def _grpc_servicer_factory(self):
        """
        Import the abstract base Servicer class from the autogenerated
        pb2_grpc module and derive a new subclass that inherits this
        Application's action objects as its interface implementation.
        """
        servicer_type_name = 'GrpcApplicationServicer'
        abstract_type = None

        # get a reference to the grpc abstract base Servicer  class
        for k, v in inspect.getmembers(self.grpc.pb2_grpc):
            if k == servicer_type_name:
                abstract_type = v
                break
        if abstract_type is None:
            raise Exception('could not find grpc Servicer class')

        # create dynamic Servicer subclass
        servicer_type = type(
            servicer_type_name, (abstract_type, ), self.actions.to_dict()
        )
        servicer = servicer_type()

        # register the Servicer with the server
        servicer.add_to_grpc_server = (
            self.grpc.pb2_grpc.add_GrpcApplicationServicer_to_server
        )

        return servicer

    def _build_pb2_modules(self):
        """
        Create the build dir inside the host package, generating the proto file
        and grpcio modules into it.
        """
        # recreate the build directory
        if os.path.isdir(self.grpc.build_dir):
            shutil.rmtree(self.grpc.build_dir)

        os.makedirs(os.path.join(self.grpc.build_dir), exist_ok=True)

        # touch the __init__.py file
        with open(os.path.join(self.grpc.build_dir, '__init__.py'), 'a'):
            pass

        # upsert the .ravel file in the build dir with a directive to ignore
        # this directory during project component discovery.
        dot_file_name = os.path.join(self.grpc.build_dir, '.ravel')
        if not os.path.isfile(dot_file_name):
            with open(dot_file_name, 'w') as dot_file:
                json.dump({'scanner': {'ignore': True}}, dot_file)
        else:
            with open(dot_file_name, 'rw') as dot_file:
                dot_data = json.load(dot_file) or {}
                dot_data.setdefault('scanner', {})['ignore'] = True
                json.dump(dot_data, dot_file)

        # generate the .proto file and generate grpc python modules from it
        self._generate_proto_file()
        self._compile_pb2_modules()

    def _generate_proto_file(self):
        """
        Iterate over function actions, using request and response schemas to
        generate protobuf message and service types.
        """
        # all lines in proto source code accumulated into "lines"
        lines = ['syntax = "proto3";']

        # service interface function delaration string accumulator
        decls = []

        # generate proto messages for each Resource class in this app
        msg_gen = MessageGenerator()
        for name, resource_type in self.res.items():
            lines.append(msg_gen.emit_resource_message(resource_type) + '\n')

        # generate request and response schemas along with service
        # interface function declarations, derived them application actions
        for action in self.actions.values():
            decls.append(action.generate_protobuf_function_declaration())
            if action.schemas.request:
                lines.append(action.generate_request_message_type())
            if action.schemas.response:
                lines.append(action.generate_response_message_type())

        # assemble service interface definition block
        lines.append('service GrpcApplication {')
        for decl in decls:
            lines.append('  ' + decl)
        lines.append('}\n')

        # write final protobuf source code to file
        console.info(
            message='generating gRPC proto file',
            data={'destination': self.grpc.proto_file}
        )
        with open(self.grpc.proto_file, 'w') as fout:
            fout.write('\n'.join(lines))

    def _compile_pb2_modules(self):
        """
        Compile the grpc .proto file, generating pb2 and pb2_grpc modules in
        the build directory. These modules contain abstract base classes
        required by the grpc server and client.
        """
        # build the shell command to run....
        protoc_command = PROTOC_COMMAND_FSTR.format(
            include_dir=os.path.realpath(
                os.path.dirname(self.grpc.proto_file)
            ) or '.',
            build_dir=self.grpc.build_dir,
            proto_file=os.path.basename(self.grpc.proto_file),
        ).strip()

        console.info(
            message='generating gRPC pb2 modules',
            data={'command': protoc_command.split('\n')}
        )

        err_msg = subprocess.getoutput(
            re.sub(r'\s+', ' ', protoc_command)
        )
        if err_msg:
            exit(err_msg)

    def _compile_grpc_web_message_classes(self, dest: Text):
        """
        Compile the grpc .proto file, generating pb2 and pb2_grpc modules in
        the build directory. These modules contain abstract base classes
        required by the grpc server and client.
        """
        # build the shell command to run....
        protoc_command = PROTOC_WEB_GENERATE_MESSAGE_CLASSES_COMMAND.format(
            include_dir=os.path.realpath(
                os.path.dirname(self.grpc.proto_file)
            ) or '.',
            build_dir=dest,
            proto_file=os.path.basename(self.grpc.proto_file),
        ).strip()

        console.info(
            message='generating gRPC-web message classes',
            data={'command': protoc_command.split('\n')}
        )
        err_msg = subprocess.getoutput(
            re.sub(r'\s+', ' ', protoc_command)
        )
        if err_msg:
            exit(err_msg)

    def _compile_grpc_web_client_stub(self, dest: Text):
        """
        Compile the grpc .proto file, generating pb2 and pb2_grpc modules in
        the build directory. These modules contain abstract base classes
        required by the grpc server and client.
        """
        # build the shell command to run....
        protoc_command = PROTOC_WEB_GENERATE_CLIENT_COMMAND.format(
            include_dir=os.path.realpath(
                os.path.dirname(self.grpc.proto_file)
            ) or '.',
            build_dir=dest,
            proto_file=os.path.basename(self.grpc.proto_file),
        ).strip()

        console.info(
            message='generating gRPC-web client stub',
            data={'command': protoc_command.split('\n')}
        )
        err_msg = subprocess.getoutput(
            re.sub(r'\s+', ' ', protoc_command)
        )
        if err_msg:
            exit(err_msg)

    def build_response_message(
        self,
        message: Message,
        source: Dict
    ) -> Message:
        """
        Recursively bind each value in a source dict to the corresponding
        attribute in a grpc Message object, resulting in the prepared response
        message, ready to be sent back to the client.
        """
        if source is None:
            return None

        for k, v in source.items():
            if v is None:
                # protobufs don't understand null values
                continue

            # TODO: move conversion logic into adapters
            if isinstance(v, (date, datetime)):
                ts = TimeUtils.to_timestamp(v)
                setattr(message, k, ts)
            elif isinstance(getattr(message, k), Message):
                sub_message = getattr(message, k)
                assert isinstance(v, dict)
                self.build_response_message(sub_message, v)
            elif isinstance(v, dict):
                json_str = self.json.encode(v)
                setattr(message, k, json_str)
            elif isinstance(v, (list, tuple, set)):
                list_field = getattr(message, k)
                sub_message_type_name = (
                    '{}Schema'.format(k.title().replace('_', ''))
                )
                sub_message_type = getattr(
                    message, sub_message_type_name, None
                )
                if sub_message_type:
                    list_field.extend(
                        self.build_response_message(sub_message_type(), v_i)
                        for v_i in v
                    )
                else:
                    v_prime = [
                        x if not isinstance(x, dict) else self.json.encode(x)
                        for x in v
                    ]
                    list_field.extend(v_prime)
            else:
                setattr(message, k, v)

        return message

    def _dump_result_obj(self, obj) -> Union[Dict, List]:
        """
        Currently, GrpcService is intended to be used for internal
        service-to-service comm. This is why we "dump" Resources here by simply
        returning their internal state dicts.
        """
        if is_resource(obj):
            return obj.internal.state
        elif is_batch(obj):
            return [x.internal.state for x in obj]
        elif isinstance(obj, (list, set, tuple)):
            return [self._dump_result_obj(x) for x in obj]
        elif isinstance(obj, dict):
            return {k: self._dump_result_obj(v) for k, v in obj.items()}
        else:
            return obj



import inspect

from inspect import Parameter
from typing import Dict, Tuple, List


class ArgumentSpecification(object):
    """
    ArgumentSpecification determines which positional and keyword arguments a
    given Guard needs. GuardMiddleware and CompositeGuard use this information
    to know which incoming action arguments should be bound to the arguments
    declared by the corresponding Guard.execute method.
    """

    def __init__(self, guard: 'Guard'):
        self.guard = guard
        self.signature = inspect.signature(guard.execute)

        # determine which arguments expected by the guard's
        # execute method that are positional and which are keyword.
        self.kwarg_keys = set()
        self.arg_keys = []
        self.arg_key_set = set()
        for idx, (k, param) in enumerate(self.signature.parameters.items()):
            if k == 'request' and (not idx):
                continue
            if param.kind != Parameter.POSITIONAL_OR_KEYWORD:
                break
            if param.default is Parameter.empty:
                self.arg_keys.append(k)
                self.arg_key_set.add(k)
            else:
                self.kwarg_keys.add(k)

        self.has_var_kwargs = False
        if 'kwargs' in self.signature.parameters:
            param = self.signature.parameters['kwargs']
            self.has_var_kwargs = param.kind == Parameter.VAR_KEYWORD

        self.has_var_args = False
        if 'args' in self.signature.parameters:
            param = self.signature.parameters['args']
            self.has_var_args = param.kind == Parameter.VAR_POSITIONAL


    def extract(self, arguments: Dict) -> Tuple[List, Dict]:
        """
        Partition arguments between a list of position arguments and a dict
        of keyword arguments.
        """
        args = [arguments[k] for k in self.arg_keys]
        if self.has_var_kwargs:
            kwargs = {
                k: v for k, v in arguments.items()
                if k not in self.arg_key_set
            }
        else:
            kwargs = {k: arguments[k] for k in self.kwarg_keys}
        return (args, kwargs)



from appyratus.schema import Schema
from appyratus.schema.fields import Field

from ravel.util import get_class_name

class FieldAdapter(object):
    def __init__(self):
        self.msg_gen = None

    def emit(self, field: Field, field_no: int):
        raise NotImplementedError()

    def bind(self, msg_gen: 'MessageGenerator'):
        self.msg_gen = msg_gen

    def emit(self, field_type, field_no, field_name, is_repeated=False):
        return '{repeated} {field_type} {field_name}{field_no}'.format(
            repeated=' repeated' if is_repeated else '',
            field_type=field_type,
            field_name=field_name,
            field_no=f' = {field_no}',
        )


class ScalarFieldAdapter(FieldAdapter):
    def __init__(self, type_name):
        self.type_name = type_name

    def emit(self, field, field_no):
        field_type = self.msg_gen.get_adapter(field).type_name
        return super().emit(
            field_type=field_type,
            field_no=field_no,
            field_name=field.name
        )


class ArrayFieldAdapter(FieldAdapter):
    def emit(self, field, field_no):
        try:
            adapter = self.msg_gen.get_adapter(field.nested)
        except:
            import ipdb; ipdb.set_trace()
        try:
            if isinstance(adapter, SchemaFieldAdapter):
                nested_field_type = get_class_name(field.nested),
            elif isinstance(adapter, NestedFieldAdapter):
                nested_field_type = get_class_name(field.nested.schema_type)
            else:
                nested_field_type = adapter.type_name
        except:
            raise Exception('Unable to establish nested field type')

        if nested_field_type.endswith('Schema'):
            nested_field_type = nested_field_type[:-len('Schema')]

        return super().emit(
            field_type=nested_field_type,
            field_no=field_no,
            field_name=field.name,
            is_repeated=True,
        )


class NestedFieldAdapter(FieldAdapter):
    def emit(self, field, field_no):
        field_type_name = get_class_name(field.schema_type)
        if field_type_name.endswith('Schema'):
            field_type_name = field_type_name[:-len('Schema')]
        return super().emit(
            field_type=field_type_name,
            field_no=field_no,
            field_name=field.name,
        )


class SchemaFieldAdapter(FieldAdapter):
    def emit(self, field, field_no, is_repeated=False):
        field_type_name = get_class_name(field)
        if field_type_name.endswith('Schema'):
            field_type_name = field_type_name[:-len('Schema')]
        return super().emit(
            field_type=field_type_name,
            field_no=field_no,
            field_name=field.name,
            is_repeated=is_repeated
        )


class EnumFieldAdapter(FieldAdapter):
    def emit(self, field, field_no):
        nested_field_type = self.msg_gen.get_adapter(field.nested).type_name
        return super().emit(
            field_type=nested_field_type,
            field_no=field_no,
            field_name=field.name,
        )



from typing import Text, List, Dict, Type, Union, Callable
from copy import deepcopy

from appyratus.utils.dict_utils import DictObject

from ravel.util.loggers import console
from ravel.query.predicate import Predicate
from ravel.util.misc_functions import (
    flatten_sequence,
    get_class_name,
)

from ravel.util import is_resource, is_resource_type
from ravel.resolver.resolver_property import ResolverProperty
from ravel.resolver.resolvers.loader import LoaderProperty

from .order_by import OrderBy
from .request import Request
from .parameters import ParameterAssignment
from .executor import Executor


class Query(object):
    def __init__(
        self,
        target: Union[Type['Resource'], Callable] = None,
        sources: List['Resource'] = None,
        parent: 'Query' = None,
        request: 'Request' = None,
        eager: bool = True,
    ):
        self.sources = sources or []
        self.target = target
        self.parent = parent
        self.options = DictObject()
        self.from_request = request
        self.eager = eager
        self.requests = {}
        self.callbacks = []
        self.parameters = DictObject(
            data={
                'where': None,
                'order_by': None,
                'offset': None,
                'limit': None,
            },
            default=None
        )

        if request:
            self.merge(request, in_place=True)

        if self.target is not None:
            self.select(self.target.ravel.schema.required_fields.keys())
            self.select(self.target.ravel.foreign_keys.keys())

    def __getattr__(self, parameter_name: str):
        return ParameterAssignment(self, parameter_name)

    def __getitem__(self, resolver: Text):
        return self.requests.get(resolver)

    def __call__(self, *args, **kwargs):
        return self.execute(*args, **kwargs)

    def __len__(self) -> int:
        return len(self.requests)

    def __contains__(self, resolver: Text) -> bool:
        return resolver in self.requests

    def __iter__(self):
        return iter(self.execute())

    def __repr__(self):
        offset = self.parameters.offest
        limit = self.parameters.limit
        return (
            f'Query('
            f'target={get_class_name(self.target)}['
            f'{offset if offset is not None else ""}'
            f':'
            f'{limit if limit is not None else ""}'
            f'])'
        )

    def merge(
        self,
        other: Union['Query', 'Request'],
        in_place: bool = False
    ) -> 'Query':
        if in_place:
            if isinstance(other, Query):
                self.parameters.update(deepcopy(other.parameters.to_dict()))
                self.options.update(deepcopy(other.options.to_dict()))
                self.select(other.requests.values())
            else:
                assert isinstance(other, Request)
                self._merge_request(other)
            return self
        else:
            merged_query = type(self)(
                target=self.target,
                parent=self.parent,
            )
            merged_query.merge(self, in_place=True)
            if isinstance(other, Query):
                merged_query.merge(other, in_place=True)
            else:
                assert isinstance(other, Request)
                merged_query._merge_request(other)
            return merged_query

    def _merge_request(self, request: 'Request'):
        # set the target type or ensure that the request's resolver
        # target type is the same as this query's, or bail.
        if request.resolver and request.resolver.target:
            if self.target is None:
                self.target = request.resolver.target
            elif self.target is not request.resolver.target:
                raise Exception(
                    'cannot merge two queries with different target types'
                )
        # marshal in raw query parameters from the request
        if request.parameters:
            params = request.parameters
            if 'select' in params:
                self.select(params.select)
            if 'where' in params:
                self.where(params.where)
            if 'order_by' in params:
                self.order_by(params.order_by)
            if 'limit' in params:
                self.limit(params.limit)
            if 'offset' in params:
                self.offset(params.offset)

    def configure(
        self,
        options: Dict = None,
        **more_options
    ) -> 'Query':
        options = dict(options or {}, **more_options)
        self.options.update(options)
        return self

    def execute(
        self,
        first=None,
        simulate=False,
    ) -> Union['Resource', 'Batch']:
        """
        Execute the query, returning a single Resource ora a Batch.
        """

        if self.eager:
            # "eager" here means that we always fetch at least ALL field
            # values by adding them to the "select" set...
            self.select(self.target.ravel.resolvers.fields.keys())

        executor = Executor(simulate=simulate)
        batch = executor.execute(self, sources=self.sources)

        if first:
            result = batch[0] if batch else None
        else:
            result = batch

        if self.callbacks:
            for func in self.callbacks:
                func(self, result)

        return result

    def exists(self):
        self.requests.clear()
        self.select(self.target._id)
        return bool(self.execute(first=True))

    def deselect(self, *args):
        """
        Remove the given arguments from the query's requests dict.
        """
        args = flatten_sequence(args)
        keys = set()

        for obj in args:
            if isinstance(obj, str):
                # if obj is str, replace it with the corresponding resolver
                # property from the target Resource class.
                _obj = getattr(self.target, obj, None)
                if _obj is None:
                    raise ValueError(f'unknown resolver: {obj}')
                obj = _obj
            elif is_resource(obj):
                keys.update(obj.internal.state.keys())
                continue
            elif is_resource_type(obj):
                keys.update(obj.ravel.resolvers.keys())
                continue

            # build a resolver request
            request = None
            if isinstance(obj, LoaderProperty) and (obj.decorator is None):
                resolver_property = obj
                keys.add(resolver_property.resolver.name)
            elif isinstance(obj, ResolverProperty):
                resolver_property = obj
                keys.add(resolver_property.resolver.name)
            elif isinstance(obj, Request):
                request = obj
                keys.add(request.resolver.name)

        for key in keys:
            if key in self.requests:
                del self.requests[key]

        return self

    def select(self, *args):
        args = flatten_sequence(args)

        for obj in args:
            if isinstance(obj, str):
                # if obj is str, replace it with the corresponding resolver
                # property from the target Resource class.
                _obj = getattr(self.target, obj, None)
                if _obj is None:
                    raise ValueError(f'unknown resolver: {obj}')
                obj = _obj
            elif is_resource(obj):
                self.select(obj.internal.state.keys())
                continue
            elif is_resource_type(obj):
                self.select(obj.ravel.resolvers.keys())
                continue

            # build a resolver request
            request = None
            if isinstance(obj, LoaderProperty) and (obj.decorator is None):
                resolver_property = obj
                request = Request(resolver_property.resolver)
            elif isinstance(obj, ResolverProperty):
                resolver_property = obj
                request = Request(resolver_property.resolver)
            elif isinstance(obj, Request):
                request = obj

            # finally set the request if one was generated
            if request:
                resolver_name = request.resolver.name
                if resolver_name not in self.target.ravel.virtual_fields:
                    if self.from_request is not None:
                        request.parent = self.from_request
                    self.requests[request.resolver.name] = request

        return self

    def where(self, *predicates, **equality_checks):
        predicates = flatten_sequence(predicates)

        for field_name, value in equality_checks.items():
            resolver_prop = getattr(self.target, field_name)
            pred = (resolver_prop == value)
            predicates.append(pred)

        if predicates:
            if self.parameters.where:
                self.parameters.where &= Predicate.reduce_and(predicates)
            else:
                self.parameters.where = Predicate.reduce_and(predicates)

        return self

    def order_by(self, *order_by):
        order_by = flatten_sequence(order_by)

        if order_by:
            self.parameters.order_by = []

            for obj in order_by:
                if isinstance(obj, OrderBy):
                    self.parameters.order_by.append(obj)
                elif isinstance(obj, str):
                    if obj.lower().endswith(' desc'):
                        order_by_obj = OrderBy(obj.split()[0], desc=True)
                    else:
                        order_by_obj = OrderBy(obj.split()[0], desc=False)
                    if order_by_obj.key not in self.target.ravel.resolvers:
                        raise ValueError(
                            f'uncognized resolver: {order_by_obj.key}'
                        )
                    self.parameters.order_by.append(order_by_obj)
        else:
            self.parameters.order_by = None

        return self

    def offset(self, offset=None):
        if offset is not None:
            self.parameters.offset = max(0, int(offset))
        else:
            self.parameters.offset = None
        return self

    def limit(self, limit):
        if limit is not None:
            self.parameters.limit = max(1, int(limit))
        else:
            self.parameters.limit = None
        return self



import inspect
import threading
import os

from threading import local, get_ident, current_thread
from concurrent.futures import (
    ThreadPoolExecutor, ProcessPoolExecutor, Future
)
from typing import (
    List, Dict, Text, Tuple, Type, Callable, Union, Optional
)
from collections import deque, OrderedDict, namedtuple

from appyratus.utils.dict_utils import DictObject, DictUtils
from appyratus.utils.string_utils import StringUtils
from appyratus.enum import EnumValueStr
from appyratus.env import Environment

from ravel.manifest.manifest import Manifest
from ravel.logging import ConsoleLoggerInterface
from ravel.util.json_encoder import JsonEncoder
from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name, inject
from ravel.constants import ID
from ravel.app.exceptions import ApplicationError
from ravel.store.transaction_manager import TransactionManager

from .action import Action
from .action_decorator import ActionDecorator
from .argument_loader import ArgumentLoader
from .resource_binder import ResourceBinder


class Mode(EnumValueStr):
    @staticmethod
    def values():
        return {
            'normal',
            'simulation',
        }


class Application(object):

    def __init__(
        self,
        middleware: List['Middleware'] = None,
        manifest: Manifest = None,
        mode: Mode = Mode.normal,
    ):
        # thread-local storage
        self.local = local()
        self.local.is_bootstrapped = False
        self.local.is_started = False
        self.local.manifest = Manifest(manifest) if manifest else None
        self.local.bootstrapper_thread_id = None
        self.local.thread_executor = None
        self.local.process_executor = None
        self.local.tx_manager = None

        self.shared = DictObject()  # storage shared by threads
        self.shared.manifest_data = None
        self.shared.app_counter = 1

        self.env = Environment()
        self._mode = mode
        self._json = JsonEncoder()
        self._binder = ResourceBinder()
        self._namespace = {}
        self._arg_loader = None
        self._logger = None
        self._root_pid = os.getpid()
        self._actions = DictObject()

        self._middleware = deque([
            m for m in (middleware or [])
            if isinstance(self, m.app_types)
        ])

        # set default main thread name
        current_thread().name = (
            f'{get_class_name(self)}MainThread'
        )


    def __getattr__(
        self,
        class_name: Text
    ) -> Optional[Union[Type['Resource'], Type['Store']]]:
        """
        This makes it possible to do app.ResourceClass or app.StoreClass as a
        convenient way of accessing registered store and resoruce classes at
        runtime--as opposed to importing them and risking annoying cyclic
        import errors.
        """
        class_obj = None
        if self.manifest is not None:
            # assume it's the name of resource class first
            class_obj = self.manifest.resource_classes.get(class_name)
            if not class_obj:
                # fall back on assuming it's a store class
                class_obj = self.manifest.store_classes.get(class_name)
        return class_obj

    def __getitem__(
        self,
        class_name: Text
    ) -> Union[Type['Resource'], Type['Store']]:
        return getattr(self, class_name, None)

    def __repr__(self):
        return (
            f'{get_class_name(self)}('
            f'bootstrapped={self.is_bootstrapped}, '
            f'started={self.is_started}, '
            f'size={len(self._actions)}'
            f')'
        )

    def __call__(
        self, *args, **kwargs
    ) ->Union[ActionDecorator, List[ActionDecorator]]:
        """
        Use this to decorate functions, adding them to this Application.
        Each time a function is decorated, it arives at the "on_decorate"
        method, where you can app the function with a web framework or
        whatever system you have in mind.

        Usage:

        ```python3
            app = Application()

            @app()
            def do_something():
                pass
        ```
        """
        if not self.is_bootstrapped:
            return self.decorator_type(self, *args, **kwargs)
        else:
            return self.start(*args, **kwargs)

    @property
    def decorator_type(self) -> Type[ActionDecorator]:
        return ActionDecorator

    @property
    def action_type(self) -> Type[Action]:
        return Action

    @property
    def manifest(self) -> 'Manifest':
        return getattr(self.local, 'manifest', None)

    @manifest.setter
    def manifest(self, manifest):
        self.local.manifest = Manifest(manifest)

    @property
    def values(self) -> Dict:
        return self.manifest.values

    @property
    def mode(self) -> Mode:
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = Mode(mode)

    @property
    def namespace(self) -> Dict:
        return self._namespace

    @property
    def is_bootstrapped(self) -> bool:
        return getattr(self.local, 'is_bootstrapped', False)

    @property
    def is_started(self) -> bool:
        return getattr(self.local, 'is_started', False)

    @property
    def is_simulation(self) -> bool:
        return self._mode == Mode.simulation

    @is_simulation.setter
    def is_simulation(self, is_simulation):
        if is_simulation:
            self._mode = Mode.simulation
        else:
            self._mode = Mode.normal

    @property
    def json(self) -> JsonEncoder:
        return self._json

    @property
    def middleware(self) -> List['Middleware']:
        return self._middleware

    @property
    def loader(self) -> 'ArgumentLoader':
        return self._arg_loader

    @property
    def actions(self) -> DictObject:
        return self._actions

    @property
    def log(self) -> ConsoleLoggerInterface:
        return self._logger

    def action(self, *args, **kwargs) -> 'ActionDecorator':
        return self.decorator_type(self, *args, **kwargs)

    def bootstrap(
        self,
        manifest: Union[Dict, Manifest, Text, Callable, 'Application'] = None,
        namespace: Dict = None,
        values: Dict = None,
        middleware: List = None,
        mode: Mode = Mode.normal,
        *args,
        **kwargs
    ) -> 'Application':
        """
        Bootstrap the data, business, and service layers, wiring them up.
        """

        def create_logger(level):
            """
            Setup root application logger.
            """
            if self.manifest.package:
                name = (
                    f'{self.manifest.package}:'
                    f'{os.getpid()}:'
                    f'{get_ident()}'
                )
            else:
                class_name = get_class_name(self)
                name = (
                    f'{StringUtils.snake(class_name)}:'
                    f'{os.getpid()}:'
                    f'{get_ident()}'
                )

            return ConsoleLoggerInterface(name, level)
            
        # warn about already being bootstrapped...
        if (
            self.is_bootstrapped and
            self.local.thread_id != get_ident()
        ):
            console.warning(
                message=f'{get_class_name(self)} already bootstrapped.',
                data={
                    'pid': os.getpid(),
                    'thread': threading.get_ident(),
                }
            )
            return self

        console.debug(f'bootstrapping {get_class_name(self)} app')

        # override the application mode set in constructor
        self._mode = Mode(mode or self.mode or Mode.normal)

        # merge additional namespace data into namespace dict
        self._namespace = DictUtils.merge(self._namespace, namespace or {})

        # register additional middleware targing this app subclass
        if middleware:
            self._middleware.extend(
                m for m in middleware if isinstance(self, m.app_types)
            )

        # build final manifest, used to bootstrap program components
        if manifest is not None:
            if isinstance(manifest, Application):
                source_app = manifest
                self.local.manifest = Manifest(source_app.shared.manifest_data)
            else:
                self.local.manifest = Manifest(manifest)
        elif self.shared.manifest_data:
            self.local.manifest = Manifest(self.shared.manifest_data)
        else:
            self.local.manifest = Manifest({})

        assert self.local.manifest is not None
        self.manifest.bootstrap(self)

        if values:
            self.manifest.values.update(values)
            
        if not self.shared.manifest_data:
            self.shared.manifest_data = self.local.manifest.data

        # set up main thread name
        if self.manifest.package:
            current_thread().name = (
                f'{StringUtils.camel(self.manifest.package)}'
            )
        else:
            # update default main thread name
            current_thread().name = (
                f'{get_class_name(self)}'
            )

        self._logger = create_logger(self.manifest.logging['level'])
        self._arg_loader = ArgumentLoader(self)

        # if we're in a new process, unset the executors so that
        # the spawn method lazily triggers the instantiation of
        # new ones at runtime.
        if not self._root_pid != os.getpid():
            self.local.thread_executor = None
            self.local.process_executor = None

        self.local.thread_id = get_ident()
        self.local.tx_manager = TransactionManager(self)

        # execute custom lifecycle hook provided by this subclass
        self.on_bootstrap(*args, **kwargs)
        self.local.is_bootstrapped = True

        console.debug(f'finished bootstrapping {get_class_name(self)} app')
        return self

    def start(self, *args, **kwargs):
        """
        Enter the main loop in whatever program context your Application is
        being used, like in a web framework or a REPL.
        """
        self.local.is_started = True
        return self.on_start(*args, **kwargs)

    def inject(
        self,
        func: Callable,
        include_resource_classes=True,
        include_store_classes=True,
        include_actions=False,
    ) -> Callable:
        """
        Inject Resource, Store, and/or Action classes into the lexical scope of
        the given function.
        """
        if include_resource_classes:
            inject(func, self.manifest.resource_classes)
        if include_store_classes:
            inject(func, self.manifest.store_classes)
        if include_actions:
            inject(func, self._actions)

        return func

    def spawn(
        self,
        target: Callable,
        args: Tuple = None,
        kwargs: Dict = None,
        multiprocessing: bool = False
    ) -> Future:
        """
        Perform a target callable in a worker thread or subprocess, creating
        and returning a Future object. This is a convenience method;
        otherwise, when you manually create a thread of process in which you
        intend to access this app, you will need to call bootstrap on the app
        again (within said thread or process context).
        """
        args = tuple(args or [])
        kwargs = kwargs or {}

        # select and lazily create the appropriate executor...
        if not multiprocessing:
            executor = getattr(self.local, 'thread_executor', None)
            if executor is None:
                thread_name_prefix = (
                    f'{StringUtils.camel(self.manifest.package)} Worker'
                    if self.manifest.package
                    else f'{get_class_name(self)} Worker'
                )
                self.local.thread_executor = ThreadPoolExecutor(
                    max_workers=os.cpu_count(),
                    initializer=self.bootstrap,
                    initargs=(self.manifest.data, ),
                    thread_name_prefix=thread_name_prefix
                )
            executor = self.local.thread_executor
        else:
            executor = getattr(self.local, 'process_executor', None)
            if executor is None:
                self.local.process_executor = ProcessPoolExecutor(
                    max_workers=os.cpu_count(),
                    initializer=self.bootstrap,
                    initargs=(self.manifest.data, ),
                )
            executor = self.local.process_executor

        # submit and return the future
        return executor.submit(target, *args, **kwargs)

    def add_action(self, action: 'Action', overwrite=False, *args, **kwargs):
        """
        Add an action to this app.
        """
        if action.name not in self._actions or overwrite:
            console.debug(f'registered action {action.name}')
            if isinstance(action, Action):
                if action.app is not self:
                    if args or kwargs:
                        console.warning(
                            '*args and **kwargs ignored by add_action'
                        )
                    decorator = self.action(
                        *action.decorator.args,
                        **action.decorator.kwargs
                    )
                    decorator(action.target)
                else:
                    self._actions[action.name] = action
            else:
                assert callable(action)
                decorator = self.action(*args, **kwargs)
                decorator(action)
        else:
            raise ApplicationError(
                message=f'action already registered: {action.name}',
                data={'action': action}
            )

    def transaction(self):
        return self.local.tx_manager

    def commit(self):
        self.local.tx_manager.commit()

    def rollback(self):
        self.local.tx_manager.rollback()

    def on_extract(self, action, index, parameter, raw_args, raw_kwargs):
        """
        Return a value for the given the name of the `inspect` module's
        parameter object. This is used by on_request when extracting action
        arguments.
        """
        if index is not None:
            return raw_args[index]

        else:
            return raw_kwargs.get(parameter.name)

    def on_bootstrap(self, *args, **kwargs):
        """
        Developer logic to execute upon calling app.bootstrap.
        """

    def on_decorate(self, action: 'Action'):
        """
        We come here whenever a function is decorated by this app. Here we
        can add the decorated function to, say, a web framework as a route.
        """

    def on_start(self, *args, **kwargs) -> object:
        """
        Custom logic to perform when app.start is called. The return value from
        on_start is used at the return value of start.
        """
        return None

    def on_request(
        self,
        action: 'Action',
        *raw_args,
        **raw_kwargs
    ) -> Tuple[Tuple, Dict]:
        """
        This executes immediately before calling a registered function. You
        must return re-packaged args and kwargs here. However, if nothing is
        returned, the raw args and kwargs are used.
        """
        args_dict = OrderedDict()
        kwargs = {}

        for idx, param in enumerate(action.signature.parameters.values()):
            if not idx:
                # idx 0 is for the Request object injected elsewhere
                continue

            target_dict = None
            index = None
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.default is inspect._empty:
                    target_dict = args_dict
                    index = idx - 1  # -1 because of injected "request" arg
                else:
                    target_dict = kwargs
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                target_dict = args_dict
                index = idx
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                target_dict = kwargs

            if target_dict is not None:
                target_dict[param.name] = self.on_extract(
                    action, index, param, raw_args, raw_kwargs
                )

        # transform positional args dict into a named tuple
        PositionalArguments = namedtuple(
            typename='PositionalArguments', field_names=args_dict.keys()
        )
        args = PositionalArguments(*tuple(args_dict.values()))

        return (args, kwargs)

    def on_response(
        self,
        action: 'Action',
        raw_result: object,
        *raw_args,
        **raw_kwargs
    ) -> object:
        """
        The return value of registered callables come here as `result`. Here
        any global post-processing can be done. Args and kwargs consists of
        whatever raw data was passed into the callable *before* on_request
        executed.
        """
        return raw_result


import asyncio
import websockets

from typing import List, Type, Dict, Tuple, Text

from ravel.util.json_encoder import JsonEncoder

from ..async_server import AsyncServer

json = JsonEncoder()


class WebsocketServer(AsyncServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._host = None
        self._port = None

    def on_bootstrap(self, host: Text, port: int):
        self._host = host
        self._port = port
        super().on_bootstrap(
            server=websockets.serve(self.serve, self.host, self.port)
        )

    def on_request(self, action, socket, request: Dict) -> Tuple[Tuple, Dict]:
        args = tuple()
        kwargs = request.get('params', {}) or {}
        kwargs['socket'] = socket
        return (args, kwargs)

    def on_response(self, action, result, *args, **kwargs):
        return json.encode(result).encode('utf-8')

    async def serve(self, socket, path):
        async for message in socket:
            print(f'>>> Processing: {message}')

            # decode raw request bytes
            try:
                request = json.decode(message)
            except ValueError:
                print(f'invalid request data: {message}')

            # route the request to the appropriate
            # action and await response
            action = self.actions.get(request['method'])
            if action is None:
                print(f'>>> Unrecognized method: {request["method"]}')
            else:
                result = await action(socket, request)
                await socket.send(result)



from typing import Text, Dict

from .constants import HTTP_METHODS, RE_HANDLER_METHOD


class FalconResource(object):
    def __init__(self, route: str = None):
        self._route = route
        self._method_2_endpoint = {}

    def __getattr__(self, key) -> 'Endpoint':
        """
        This simulates falcon resource methods, like on_get,
        on_post, etc, when called on the instance.
        """
        match = RE_HANDLER_METHOD.match(key)
        if match is not None:
            method = match.groups()[0].upper()
            if not self.is_method_supported(method):
                raise AttributeError(
                    f'HTTP {method.upper()} unsupported '
                    f'for URL path {self._route}'
                )
            endpoint = self.get_endpoint(method)
            return endpoint

        raise AttributeError(key)

    @property
    def route(self) -> str:
        return self._route

    def get_method_map(self) -> Dict[Text, 'Endpoint']:
        return self._method_2_endpoint.copy()

    def get_endpoint(self, method: Text) -> 'Endpoint':
        return self._method_2_endpoint.get(method.upper())

    def add_endpoint(self, endpoint: 'Endpoint'):
        # TODO: Raise exception regarding already registered to http method
        self._method_2_endpoint[endpoint.method.upper()] = endpoint

    def is_method_supported(self, method: Text) -> bool:
        is_recognized = method.upper() in HTTP_METHODS
        is_registered = method.upper() in self._method_2_endpoint
        return (is_recognized and is_registered)



from typing import Dict, Text, Type

from ravel.util import get_class_name
from ravel.util.json_encoder import JsonEncoder
from ravel.schema import Schema, fields as field_types

STRING_FIELD_TYPE_2_FORMAT = {
    field_types.UuidString: 'uuid',
    field_types.Email: 'email',
    field_types.Url: 'uri',
    field_types.DateTimeString: 'date-time',
}


class JsonSchemaGenerator:
    def __init__(self, hostname: Text = None):
        self.hostname = (hostname or '').rstrip('/')
        self.json = JsonEncoder()

    def from_schema(self, schema: 'Schema', encode=False) -> Dict:
        schema_class_name = self.get_schema_name(schema)
        json_schema = {
            '$schema': 'http://json-schema.org/schema#',
            '$id': f'{self.hostname}/schemas/{schema_class_name}.json',
            'properties': self._derive_properties(schema),
            'required': list(schema.required_fields.keys()),
        }
        if encode:
            return self.json.encode(json_schema)
        return json_schema

    def from_resource(self, resource_type: Type['Resource'], encode=False) -> Dict:
        json_schema = self.from_schema(resource_type.ravel.schema)
        for resolver in resource_type.ravel.resolvers.values():
            if resolver.private:
                continue
            if resolver.name not in resource_type.ravel.schema.fields:
                target_schema = resolver.target.ravel.schema
                field_dict = self._derive_property(target_schema)
                if resolver.many:
                    json_schema['properties'][resolver.name] = {
                        'type': 'array',
                        'default': [],
                        'items': field_dict
                    }
                else:
                    json_schema[resolver.name] = field_dict
        if encode:
            return self.json.encode(json_schema)
        return json_schema

    def _derive_properties(self, schema):
        properties = {}
        for field in schema.fields.values():
            if not field.meta.get('private', False):
                prop = self._derive_property(field)
                properties[field.name] = prop
        return properties

    def _derive_property(self, field: 'Field'):
        field_dict = {}
        if isinstance(field, field_types.String):
            field_dict['type'] = 'string'
            format_str = STRING_FIELD_TYPE_2_FORMAT.get(type(field))
            if format_str is not None:
                field_dict['format'] = format_str
        elif isinstance(field, field_types.Bool):
            field_dict['type'] = 'boolean'
        elif isinstance(field, field_types.Int):
            field_dict['type'] = 'integer'
        elif isinstance(field, field_types.Float):
            field_dict['type'] = 'number'
        elif isinstance(field, field_types.DateTime):
            field_dict['type'] = 'integer'
            field_dict['format'] = 'date-time'
        elif isinstance(field, field_types.DateTimeString):
            field_dict['type'] = 'string'
            field_dict['format'] = 'date-time'
        elif isinstance(field, field_types.Dict):
            field_dict['type'] = 'object'
        elif isinstance(field, (field_types.List, field_types.Set)):
            field_dict['type'] = 'array'
            field_dict['items'] = self._derive_property(field.nested)
            field_dict['default'] = []
        elif isinstance(field, field_types.Enum):
            inner_field_dict = self._derive_property(field.nested)
            field_dict['type'] = inner_field_dict['type']
            field_dict['enum'] = list(field.values)
        elif isinstance(field, Schema):
            defined_name = self.get_schema_name(field)
            field_dict['$ref'] = defined_name
        elif isinstance(field, field_types.Nested):
            defined_name = self.get_schema_name(field.schema_type)
            field_dict['$ref'] = defined_name
        elif isinstance(field, field_types.Uuid):
            field_dict['type'] = 'string'
            field_dict['format'] = 'uuid'
        else:
            field_dict['type'] = 'null'

        if field.nullable:
            field_dict['type'] = [field_dict['type'], 'null']

        return field_dict

    @staticmethod
    def get_schema_name(schema: 'Schema') -> Text:
        name = get_class_name(schema)
        if name.endswith('Schema'):
            name = name[:-6]
        return name




from typing import Text, Dict

from ravel.schema import Schema, fields


class Binding:

    class Schema(Schema):
        resource = fields.String()
        store = fields.String()
        params = fields.Dict(default={})


    def __init__(
        self,
        resource_class_name: Text,
        store_class_name: Text,
        bind_params: Dict = None
    ):
        self.resource_class_name = resource_class_name
        self.store_class_name = store_class_name
        self.bind_params = bind_params or {}
        self.resource_class = None
        self.store_class = None


from typing import Dict, List, Text, Set

from ravel import Resource, Relationship, fields
from ravel.constants import ID, REV

# TODO: implement updates to depth in insert and unlink

class RecursiveList(Resource):
    """
    # RecursiveList
    Each RecursiveList is essentially a linked list, where node that has a
    String name. In addition, each node may have "child" RecursiveLists, in which
    case it is called a "parent" . In other words, this is a recursive data
    structure.

    Let's look at an example. Suppose we have siblings (A, B, C) which are all
    children of X. Then the following statements are true of its relationships:

    ```python
    assert {X} == {A.parent, B.parent, C.parent}
    assert {A, B, C} == set(X.children)
    assert A == B.previous
    assert C == B.following
    ```

    ## Fields
    - `name`: String identifier for this RecursiveList
    - `size`: The number of its children RecursiveLists
    - `position`: The positional index of this RecursiveList in its parent
    """

    '''
    size = fields.Int(nullable=False, default=lambda: 0, required=True)
    depth = fields.Int(nullable=False, default=lambda: 0, required=True)
    position = fields.Int(nullable=False, required=True, default=lambda: -1, private=True)
    parent_id = fields.String(required=True, default=lambda: None, private=True)

    parent = Relationship(
        join=lambda cls: (cls.parent_id, cls._id),
        immutable=True,
    )

    children = Relationship(
        join=lambda cls: (cls._id, cls.parent_id),
        #order_by=lambda cls: cls.position.asc,
        immutable=True,
        many=True,
    )

    previous = Relationship(
        join=lambda cls: cls,
        where=lambda sources: (
            (sources.resource_type.parent_id == sources[0].parent_id) &
            (sources.resource_type.position == sources[0].position - 1)
        ),
        immutable=True,
    )

    following = Relationship(
        join=lambda cls: cls,
        where=lambda sources: (
            (sources.resource_type.parent_id == sources[0].parent_id) &
            (sources.resource_type.position == sources[0].position + 1)
        ),
        immutable=True,
    )


    @classmethod
    def __abstract__(cls):
        return True

    @property
    def has_parent(self):
        return self.parent_id is not None

    def copy(self, fields: Set[Text] = None, recursive=False):
        def copy_one(node, parent_id, fields):
            return self.__class__(
                data={k: node.internal.state[k] for k in fields},
                parent_id=parent_id,
            ).create()

        def copy_children(node, parent_id, fields):
            if not node.size:
                return
            copied_children = self.Batch([])
            node.children.resolve(fields)
            for child in node.children:
                copied_child = copy_one(child, parent_id, fields)
                copied_children.append(copied_child)
            copied_children.create()
            for child, copied_child in zip(node.children, copied_children):
                copy_children(child, copied_child._id, fields)

        # copy all non-internal fields by default
        fields = (
            set(fields or self.schema.fields.keys())
                - {ID, REV}
        )

        # copy this instance & conditionally children, recursively
        copied = copy_one(self, self.parent_id, fields)
        if recursive:
            copy_children(self, copied._id, fields)

        return copied

    def unlink(self):
        """
        Disassociate this RecursiveList with its parent, adjusting the indexes
        of all sibling RecursiveLists that remain.
        """
        if self.parent_id is None:
            return

        dirty = self.Batch([self, self.parent])  # resources to update

        # adjust remaining sibling's positions
        siblings = self.query(
            (self.__class__.parent_id == self._id) &
            (self.__class__.position > self.position)
        )
        if siblings:
            dirty.extend(siblings)
            for sibling in siblings:
                sibling.position -= 1

        self.position = -1
        self.parent_id = None
        self.parent.size -= 1
        self.unload({'parent', 'previous', 'following'})
        dirty.update()

    def insert_many(self, children: List['RecursiveList'], index: int, copy=False):
        """
        Remove each child RecursiveList from whatever list it may already belong to
        and insert it in this RecursiveList, adjusting the position of all sibling
        RecursiveLists downstream.
        """
        position = max(0, min(self.size, index))
        do_shift_sibling_positions = position < self.size

        dirty = self.Batch([self])  # resources to update

        for j, child in enumerate(children):
            if copy:
                child = child.copy(recursive=True)
            elif child.has_parent:
                child.unlink()
            child.unload({'parent', 'following', 'previous'})
            child.parent_id = self._id
            child.position = position + j
            dirty.append(child)
            self.size += 1

        if do_shift_sibling_positions:
            siblings = self.query(
                predicate=(
                    (self.__class__.parent_id == self._id) &
                    (self.__class__.position >= position)
                ),
                order_by=self.__class__.position.asc
            )
            if siblings:
                dirty.extend(siblings)
                offset = position + len(children)
                for i, sibling in enumerate(siblings):
                    sibling.position = offset + i

        self.unload('children')
        dirty.update()

    def insert(self, child: 'RecursiveList', index: int, copy=False):
        """
        Insert a RecursiveList at the given index.
        """
        self.insert_many([child], index, copy=copy)

    def unshift(self, child: 'RecursiveList', copy=False):
        """
        Insert a child RecursiveList at the head of this RecursiveList's children.
        """
        self.insert_many([child], 0, copy=copy)

    def unshift_many(self, children: List['RecursiveList'], copy=False):
        """
        Insert a list of RecursiveLists at the head of this RecursiveList's children.
        """
        self.insert_many(children, 0, copy=copy)

    def push(self, child: 'RecursiveList', copy=False):
        """
        Insert a child RecursiveList at the tail of this RecursiveList's children.
        """
        self.insert_many([child], self.size, copy=copy)

    def push_many(self, children: List['RecursiveList'], copy=False):
        """
        Insert a list of RecursiveLists at the tail of this RecursiveList's children.
        """
        self.insert_many(children, self.size, copy=copy)

    def reverse(self):
        for i, child in enumerate(self.children):
            child.position = len(self.children) - i - 1
            child.unload({'following', 'previous'})
        self.unload('children')
        self.children.update()

    def apply(self, func, direction, args=None, kwargs=None, inclusive=True):
        args = args or tuple()
        kwargs = kwargs or {}
        results = []
        if inclusive:
            result = func(self, *args, **kwargs)
            results.append(result)
        if direction < 0:
            node = self.r.parent.query()
            while node is not None:
                result = func(node, *args, **kwargs)
                results.append(result)
                node = node.parent
        else:
            for node in self.r.children.query():
                child_results = node.apply(func, direction, args, kwargs)
                results.extend(child_results)
        return results

'''












if __name__ == '__main__':
    from ravel.app import ReplApplication

    repl = ReplApplication()

    class RecursiveList(RecursiveList):
        name = fields.String()

    @repl()
    def test_push():
        parent = RecursiveList(name='parent', position=0).save()
        for i in range(3):
            child = RecursiveList(name=f'child_{i}').create()
            parent.push(child)

        for s1, s2 in zip(parent.children, parent.children[1:]):
            assert s1.position == s2.position - 1
            assert int(s1.name[-1]) == int(s2.name[-1]) - 1

        repl.namespace.update(locals())

    @repl()
    def test_unshift():
        parent = RecursiveList(name='parent', position=0).save()
        for i in range(3):
            child = RecursiveList(name=f'child_{i}').create()
            parent.unshift(child)

        for s1, s2 in zip(parent.children, parent.children[1:]):
            assert s1.position == s2.position - 1
            assert int(s1.name[-1]) == int(s2.name[-1]) + 1

        repl.namespace.update(locals())

    @repl()
    def test_unlink():
        parent = RecursiveList(name='parent', position=0).save()
        child = RecursiveList(name='child').save()

        parent.unshift(child)

        assert parent.size == 1
        assert parent._id == child.parent_id
        assert child.position == 0

        child.unlink()

        assert parent.size == 0
        assert child.parent_id is None
        assert child.position == -1

        repl.namespace.update(locals())

    @repl()
    def test_reverse():
        parent = RecursiveList(name='parent', position=0).save()
        for i in range(3):
            child = RecursiveList(name=f'child_{i}').create()
            parent.push(child)

        parent.reverse()

        for s1, s2 in zip(parent.children, parent.children[1:]):
            assert int(s1.name[-1]) == int(s2.name[-1]) - 1

        repl.namespace.update(locals())

    @repl()
    def test_insert_many():
        parent = RecursiveList(name='parent', position=0).save()
        children = [
            RecursiveList(name=f'child_{i}').create() for i in range(1, 6)
        ]
        for child in children[:2]:
            parent.push(child)

        parent.insert_many(children[2:], 1, copy=True)

        for s1, s2 in zip(parent.children[1:], parent.children[2:]):
            assert s1.position == s2.position - 1
            if s2 is not parent.children[-1]:
                assert int(s1.name[-1]) == int(s2.name[-1]) - 1

        assert parent.children[-1].name == 'child_2'

        repl.namespace.update(locals())


    repl.bootstrap(namespace=globals()).start()



from typing import Dict, Tuple

from appyratus.utils.dict_utils import DictObject

from ravel.exceptions import NotAuthorized
from ravel.util.misc_functions import is_sequence, normalize_to_tuple
from ravel.app.middleware.middleware import Middleware

from .guard import Guard, GuardFailure


class EvaluateGuards(Middleware):
    """
    Evaluate the guard/guards provided through an action decorator's `guard`
    or `guards` kwarg. If any guard returns False, raise an exception;
    however, if a guard raises its own exception, this middleware allows it
    to bubble up.
    """

    def on_request(
        self,
        action: 'Action',
        request: 'Request',
        processed_args: Tuple,
        processed_kwargs: Dict,
    ):
        """
        If the requested action has a guard, we execute it here. If it fails,
        we raise a GuardFailure exception.
        """
        guard = self._resolve_guard(action)
        if guard is not None:
            # merge all positional and keyword args into a single dict
            arguments = self._extract_args(
                action, processed_args, processed_kwargs
            )
            # execute the guard (and composite guards, recursively)
            guard_passed = guard(request, arguments=arguments)
            if guard_passed is False:
                raise GuardFailure(guard, "guard failed")

    def _resolve_guard(self, action):
        """
        Return the guard associated with the action if it exists. The argument
        is expected to be a Guard instance or a callback function that returns
        one.
        """
        guard = action.guard
        if not guard:
            return None
        elif not isinstance(guard, Guard) and callable(guard):
            guard = guard()
        return guard

    def _extract_args(self, action, args, kwargs) -> Dict:
        """
        Merge all args and kwargs into a single Dict. This is is passed into
        each executed Guard as **kwargs.
        """
        arg_names = [k for k in action.signature.parameters][:len(args)]
        merged = dict(zip(arg_names, args))
        merged.update(kwargs)
        return merged



import inspect
import traceback
import logging

from inspect import Signature
from typing import Dict, Text, Tuple, Callable
from collections import deque

from appyratus.utils.time_utils import TimeUtils
from appyratus.utils.dict_utils import DictObject

from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name, get_callable_name

from ravel.exceptions import RavelError


class ActionError(RavelError):
    """
    An `Error` simply keeps a record of an Exception that occurs and the
    middleware object which raised it, provided there was one. This is used
    in the processing of requests, where the existence of errors has
    implications on control flow for middleware.
    """

    def __init__(self, exc, middleware=None):
        self.middleware = middleware
        self.timestamp = TimeUtils.utc_now()
        self.exc = exc
        if isinstance(exc, RavelError) and exc.wrapped_traceback:
            self.trace = self.exc.wrapped_traceback.strip().split('\n')[1:]
            self.exc_message = self.trace[-1].split(': ', 1)[1]
        else:
            self.trace = traceback.format_exc().strip().split('\n')[1:]
            final_line = self.trace[-1]
            if ':' in final_line:
                self.exc_message = final_line.split(':', 1)[1]
            else:
                self.exc_message = None

    def to_dict(self):
        data = {
            'timestamp': self.timestamp.isoformat(),
            'message': self.exc_message,
            'traceback': self.trace,
        }

        if isinstance(self.exc, RavelError):
            if not self.exc.wrapped_exception:
                data['exception'] = get_class_name(self.exc)
            else:
                data['exception'] = get_class_name(self.exc.wrapped_exception)

            if self.exc.logged_traceback_depth is not None:
                depth = self.exc.logged_traceback_depth
                data['traceback'] = data['traceback'][-2*depth:]

        if self.middleware is not None:
            data['middleware'] = get_class_name(self.middleware)

        if isinstance(self.exc, RavelError):
            data.update(self.exc.data)

        return data


class BadRequest(RavelError):
    """
    If any exceptions are raised during the execution of Middleware or an
    action callable itself, we raise BadRequest.
    """

    def __init__(self,
        action: 'Action',
        state: 'ExecutionState',
        *args,
        **kwargs
    ):
        super().__init__(
            f'error/s occured in action '
            f'"{action.name}" (see logs)'
        )

        self.action = action
        self.state = state


class ExecutionState(object):
    def __init__(self, action, raw_args, raw_kwargs):
        self.action = action
        self.errors = []
        self.exc = None
        self.middleware = []
        self.raw_args = raw_args
        self.raw_kwargs = raw_kwargs
        self.processed_args = None
        self.processed_kwargs = None
        self.raw_result = None
        self.result = None
        self.is_complete = False

    @property
    def app(self) -> 'Application':
        return self.action.app


class Request(object):
    def __init__(self, state):
        self.__dict__['internal'] = state
        self.__dict__['context'] = DictObject()

    def __repr__(self):
        return f'Request(action="{self.internal.action.name}")'

    def __getattr__(self, key):
        return self.context.get(key)

    @property
    def is_complete(self) -> bool:
        return self.internal.is_complete

    @is_complete.setter
    def is_complete(self, is_complete: bool):
        self.internal.is_complete = is_complete

    @property
    def raw_args(self):
        return self.internal.raw_args

    @property
    def raw_kwargs(self):
        return self.internal.raw_kwargs

    @property
    def session(self) -> 'Session':
        return self.context.session

    @session.setter
    def session(self, session: 'Session'):
        self.context.session = session

    @property
    def app(self) -> 'Application':
        return self.internal.app


class Action(object):

    Error = ActionError
    State = ExecutionState
    BadRequest = BadRequest

    def __init__(self, func: Callable, decorator: 'ActionDecorator'):
        self._is_bootstrapped = False
        self._decorator = decorator
        self._target = func.target if isinstance(func, Action) else func
        self._signature = inspect.signature(self._target)
        self._api_object = decorator.api_object

    def __repr__(self):
        return f'{get_class_name(self)}(name={self.name})'

    def __getattr__(self, key: Text):
        return self.decorator.kwargs.get(key)

    def __call__(self, *raw_args, **raw_kwargs):
        state = ExecutionState(self, raw_args, raw_kwargs)
        request = Request(state)

        for func in (
            self._apply_middleware_pre_request,
            self._apply_app_on_request,
            self._apply_middleware_on_request,
            self._apply_target_callable,
            self._apply_app_on_response,
        ):
            error = func(request)
            if (error is not None) or request.is_complete:
                break

        if state.exc is None:
            self._apply_middleware_post_request(request)
        else:
            self._apply_middleware_post_bad_request(request)

        if state.errors:
            console.error(
                message=f'error/s occured in action: {self.name}',
                data={
                    'errors': [
                        error.to_dict() for error in state.errors
                    ]
                }
            )
            raise BadRequest(self, state)

        return state.result

    @property
    def target(self):
        return self._target

    @property
    def decorator(self):
        return self._decorator

    @property
    def app(self) -> 'Application':
        return self._decorator.app

    @property
    def api_object(self) -> 'Api':
        # TODO: rename this thing
        return self._api_object

    @property
    def is_bootstrapped(self) -> bool:
        return self._is_bootstrapped

    @property
    def name(self) -> Text:
        return get_callable_name(self._target)

    @property
    def docstring(self) -> Text:
        return inspect.getdoc(self._target)

    @property
    def signature(self) -> Signature:
        return self._signature

    @property
    def source(self) -> Text:
        return inspect.getsource(self.target)

    def bootstrap(self):
        console.debug(f'bootstrapping {self.name} action')
        self._prepare_middleware()
        self.on_bootstrap()
        self._is_bootstrapped = True

    def on_bootstrap(self):
        pass

    def _prepare_middleware(self):
        """
        Determine which middleware hook methods will or will potentially be
        invoked during the execution of this action. This is computed on
        bootstrap.
        """
        self._mware_pre_request_methods = deque()
        self._mware_on_request_methods = deque()
        self._mware_post_request_methods = deque()
        self._mware_post_bad_request_methods = deque()

        def is_virtual(func):
            return getattr(func, 'is_virtual', False)

        for mware in self.app.middleware:
            if not isinstance(self.app, mware.app_types):
                continue

            if not is_virtual(mware.pre_request):
                self._mware_pre_request_methods.append(mware.pre_request)
            if not is_virtual(mware.on_request):
                self._mware_on_request_methods.append(mware.on_request)
            if not is_virtual(mware.post_request):
                self._mware_post_request_methods.appendleft(mware.post_request)
            if not is_virtual(mware.post_bad_request):
                self._mware_post_bad_request_methods.appendleft(mware.post_bad_request)

    def _apply_middleware_pre_request(self, request):
        state = request.internal
        error = None

        for func in self._mware_pre_request_methods:
            try:
                func(self, request, state.raw_args, state.raw_kwargs)
                state.middleware.append(func)
            except Exception as exc:
                error = ActionError(exc, func)
                state.errors.append(error)
                break

            # if is_complete, we abort further mware processing as well as all
            # subsequent steps in processing the action. instead, we skip
            # directly to post_request middleware execution.
            if request.is_complete:
                break

        return error

    def _apply_middleware_on_request(self, request):
        state = request.internal
        error = None

        for func in self._mware_on_request_methods:
            args = state.processed_args
            kwargs = state.processed_kwargs

            try:
                func(self, request, args, kwargs)
            except Exception as exc:
                error = Action.Error(exc, mware)
                state.errors.append(error)

        return error

    def _apply_middleware_post_request(self, request):
        state = request.internal
        error = None

        for func in self._mware_post_request_methods:
            try:
                func(self, request, state.result)
            except Exception as exc:
                error = Action.Error(exc, func)
                state.errors.append(error)

        return error

    def _apply_middleware_post_bad_request(self, request):
        state = request.internal
        error = None

        for func in self._mware_post_bad_request_methods:
            try:
                func(self, request, state.exc.exc)
            except Exception as exc:
                error = Action.Error(exc, mware)
                state.errors.append(error)

        return error

    def _apply_target_callable(self, request):
        state = request.internal
        try:
            error = None
            takes_arguments = len(self.signature.parameters) > 1
            if takes_arguments:
                state.raw_result = self._target(
                    request,
                    *state.processed_args,
                    **state.processed_kwargs
                )
            else:
                state.raw_result = self._target(request)
        except Exception as exc:
            error = Action.Error(exc)
            state.exc = error
            state.errors.append(error)

        return error

    def _apply_app_on_response(self, request):
        state = request.internal
        error = None
        try:
            state.result = self.decorator.app.on_response(
                self,
                state.raw_result,
                state.raw_args,
                state.raw_kwargs,
                *state.processed_args,
                **state.processed_kwargs
            )
        except Exception as exc:
            error = Action.Error(exc)
            state.errors.append(error)

        return error

    def _apply_app_on_request(self, request):
        state = request.internal
        error = None
        try:
            if self._api_object is not None:
                params = self.app.on_request(
                    self, self._api_object, *state.raw_args, **state.raw_kwargs
                )
            else:
                params = self.app.on_request(
                    self, *state.raw_args, **state.raw_kwargs
                )
            args, kwargs = (
                params if params
                else (state.raw_args, state.raw_kwargs)
            )
            state.processed_args, state.processed_kwargs = (
                self.marshal(args, kwargs)
            )
        except Exception as exc:
            error = Action.Error(exc)
            state.errors.append(error)

        return error

    def marshal(self, raw_args, raw_kwargs):
        return self.app.loader.load(self, raw_args, raw_kwargs)


class AsyncAction(Action):
    """
    This specialized `Action` can be used by any new `Application` type
    whose wrapped functions are coroutines.
    """
    # TODO: Implement



from typing import Dict, Union

from ravel.schema import Schema, fields
from ravel.util.misc_functions import get_callable_name
from ravel.resolver.resolver import Resolver
from ravel.resolver.resolvers.loader import Loader, View
from ravel.resolver.resolvers.relationship import Relationship


resolver = Resolver.build_decorator()
relationship = Relationship.build_decorator()
view = View.build_decorator()


class field:
    """
    This decorator is NOT a ResolverDecorator, but it provides a way that
    fields can be defined on a Resource class while simultaneously defining
    custom resolver logic for loading its value. All uses of this decorator
    will result in a new Field being added to the resource Schema, using the
    name of the decorated method as the name of the field.
    """

    def __init__(self, field: 'Field'):
        self.field = field

    def __call__(self, func):
        self.field.meta['ravel_on_resolve'] = func
        self.field.name = self.field.source = get_callable_name(func)
        return self.field


class nested(view):
    def __init__(self, schema: Union[Dict, Schema], *args, **kwargs):
        super().__init__(field=fields.Nested(schema), *args, **kwargs)



from ravel import Resource, fields


class Thing(Resource):
    anything = fields.Field()
    name = fields.String()
    option = fields.Enum(fields.String(), {'a', 'b', 'c'})
    code = fields.Bytes()
    ident = fields.UuidString()
    when = fields.DateTimeString()
    age = fields.Int()
    email = fields.Email()
    real = fields.Float()
    key = fields.Uuid()
    happy = fields.Bool()
    dob = fields.DateTime()
    timestamp = fields.Timestamp()
    colors = fields.List(fields.String())
    reals = fields.List(fields.Float())
    integers = fields.Set(fields.Int())
    blob = fields.Dict()
    nested = fields.Nested({
        'email': fields.Email(),
        'substrings': fields.List(fields.String()),
        'subdicts': fields.List(fields.Nested({
            'age': fields.Int(),
            'name': fields.String(),
        }))
    })



from appyratus.enum import Enum

from ravel.schema import Field
from ravel.resolver.resolvers.loader import Loader, LoaderProperty
from ravel.query.predicate import Predicate, ConditionalPredicate


SUPPORTED_GEOMETRY_TYPES = Enum(
    POINT='POINT',
    POLYGON='POLYGON',
)

POSTGIS_OP_CODE = Enum(
    CONTAINS='contains',
    CONTAINED_BY='contained-by',
    WITHIN_RADIUS='within-radius',
)


class GeometryObject:
    def __init__(self, geometry_type):
        self.geometry_type = geometry_type

    def to_EWKT_string(self) -> str:
        raise NotImplementedError('override in subclass')


class PointGeometry(GeometryObject):
    def __init__(self, obj):
        if isinstance(obj, PointGeometry):
            vertex = obj.vertex
        elif isinstance(obj, dict):
            vertex = obj['vertex']
        else:
            vertex = obj

        super().__init__(SUPPORTED_GEOMETRY_TYPES.POINT)
        self.vertex = vertex

    def __getitem__(self, idx):
        return self.vertex[idx]

    def to_EWKT_string(self) -> str:
        x, y = self.vertex
        return f'SRID=4326; POINT({x} {y})'


class PolygonGeometry(GeometryObject):
    def __init__(self, obj):
        if isinstance(obj, PolygonGeometry):
            vertices = obj.vertices
        elif isinstance(obj, dict):
            vertices = obj['vertices']
        else:
            vertices = obj

        super().__init__(SUPPORTED_GEOMETRY_TYPES.POLYGON)
        self.vertices = vertices

    def to_EWKT_string(self) -> str:
        points = [f'{v[0]} {v[1]}' for v in self.vertices]
        return f'SRID=4326; POLYGON(({ ",".join(points) }))'


class PostgisGeometryLoaderProperty(LoaderProperty):
    def contains(self, geometry: 'GeometryObject') -> Predicate:
        return ConditionalPredicate(POSTGIS_OP_CODE.CONTAINS, self, geometry)

    def is_within_radius_of(self, point, radius):
        value = {'center': PointGeometry(point).vertex, 'radius': radius}
        return ConditionalPredicate(
            POSTGIS_OP_CODE.WITHIN_RADIUS, self, value,
            ignore_field_adapter=True
        )

    def is_contained_by(self, polygon: 'PolygonGeometry') -> Predicate:
        return ConditionalPredicate(
            POSTGIS_OP_CODE.CONTAINED_BY, self, polygon
        )


class PostgisGeometryLoader(Loader):

    @classmethod
    def property_type(cls):
        return PostgisGeometryLoaderProperty


class Geometry(Field):
    def __init__(self, geo_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta['resolver_type'] = PostgisGeometryLoader
        self.geo_type = geo_type


class Point(Geometry):
    def __init__(self, *args, **kwargs):
        super().__init__(SUPPORTED_GEOMETRY_TYPES.POINT, *args, **kwargs)

    def process(self, obj):
        if isinstance(obj, list):
            if len(obj) != 2:
                return (None, 'expected list with 2 numbers')
            point = PointGeometry(obj)
            return (point, None)

        if isinstance(obj, dict):
            assert obj.get('type') == SUPPORTED_GEOMETRY_TYPES.POINT
            point = PointGeometry(obj['vertex'])
            return (point, None)

        if not isinstance(obj, PointGeometry):
            return (None, 'expected a PointGeometry object')

        return (obj, None)


class Polygon(Geometry):
    """
    A Schema field representing a Postgis Geometry, represented by a
    GeometryObject. The SqlalchemyStore corresponding field adapter takes care
    of transforming the dict to and from the EWKT strings understood by
    Postgis.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(SUPPORTED_GEOMETRY_TYPES.POLYGON, *args, **kwargs)

    def process(self, obj):
        if isinstance(obj, list):
            polygon = PolygonGeometry(obj)
            return (polygon, None)

        if isinstance(obj, dict):
            geo_type = obj.get('type')
            if geo_type != SUPPORTED_GEOMETRY_TYPES.POLYGON:
                return (None, 'expected a PolygonGeometry object')

            vertices = obj.get('vertices') or []
            if not isinstance(vertices, (tuple, list)):
                return (None, f'unrecognized vertices format: {vertices}')

            polygon = PolygonGeometry(vertices)
            return (polygon, None)

        if not isinstance(obj, PolygonGeometry):
            return (None, 'expected a Polygon object')

        return (obj, None)



from inspect import Parameter
from collections import defaultdict
from typing import (
    List, Dict, ForwardRef, Text, Tuple, Set, Type, Callable,
    _GenericAlias as GenericAlias
)

from ravel.constants import ID, REV
from ravel.util.misc_functions import (
    extract_res_info_from_annotation,
    is_sequence,
)

from ravel.util import (
    is_resource
)

class ArgumentLoader(object):
    """
    A `ArgumentLoader` instance is used by each `Application` object. When enabled
    (which it is, by default), arguments passed into API actions (AKA Action
    objects) are automatically loaded or converted to their corresponding
    Resources, replacing the raw arguments, provided that the arguments are
    declared using a Resource as the type annotation.

    For example, if you have an API action, like:

    ```python3
    @repl()
    def get_projects(user: User) -> Project.Batch:
        return user.projects
    ```

    Then the `ArgumentLoader` makes it possible to call this method in the
    following ways:

    ```python3
    get_projects(user)
    get_projects("user-id-123")
    get_projects({"id": "user-id-124", "name": "Musk"})
    ```

    In the first case, a `User` Resource is passed in. In the second, a User ID
    is passed in, and the argument loader performs `User.get(_id)`. Finally, in
    the third case, a raw dict is passed in with the format specified by the
    `User` Resource schema and is converted into the corresponding Resource.
    """

    class ArgumentSpec(object):
        def __init__(
            self,
            position: int,
            arg_name: Text,
            many: bool,
            resource_type: Type['Resource']
        ):
            self.position = position
            self.arg_name = arg_name
            self.many = many
            self.resource_type = resource_type

    def __init__(self, app: 'Application', on_load: Callable = None):
        self._app = app
        self._on_load = (
            on_load or (lambda spec, raw_value, loaded_value: loaded_value)
        )
        self._action_2_specs = defaultdict(list)
        self.bind()

    def bind(self):
        self._action_2_specs = defaultdict(list)
        for action in self._app.actions.values():
            self._action_2_specs[action] = []
            for idx, param in enumerate(action.signature.parameters.values()):
                ann = param.annotation
                many, res_class_name = extract_res_info_from_annotation(ann)
                if res_class_name is None:
                    continue
                resource_type = self._app[res_class_name]
                if resource_type is not None:
                    position = (
                        idx if param.default == Parameter.empty else None
                    )
                    spec = ArgumentLoader.ArgumentSpec(
                        idx, param.name, many, resource_type
                    )
                    self._action_2_specs[action].append(spec)

    def load(self, action: 'Action', args: Tuple, kwargs: Dict) -> Tuple:
        """
        Replace args and kwargs with corresponding Entity and return them
        """
        loaded_args = list(args)
        loaded_kwargs = kwargs.copy()

        for spec in self._action_2_specs[action]:
            pos = spec.position
            posindex = pos-1
            if pos is not None and pos <= len(args):
                raw_arg_value = args[posindex]
                loaded_args_or_kwargs = loaded_args
                key = posindex
                is_positional = True
            else:
                raw_arg_value = kwargs.get(spec.arg_name)
                loaded_args_or_kwargs = loaded_kwargs
                key = spec.arg_name
                is_positional = False

            # Note that "key" is either a position integer offset
            # of the name of a keyword argument.

            loaded_entity = self.load_param(
                spec.many, spec.resource_type, raw_arg_value
            )
            # store a reference to the raw argument value on the loaded Entity
            # so that it can still be accessed inside the app.
            if loaded_entity is not None:
                loaded_entity.internal.arg = raw_arg_value

            loaded_entity = self._on_load(
                spec, raw_arg_value, loaded_entity
            )

            if is_positional:
                loaded_args[key] = loaded_entity
            else:
                loaded_kwargs[key] = loaded_entity

        return (tuple(loaded_args), loaded_kwargs)

    def load_param(self, many: bool, resource_type: Type['Resource'], preloaded):
        """
        Convert the given parameter "preloaded" into its corresponding
        Entity.

        - If the preloaded value is an ID, fetch the object.
        - If it is a list of IDs, return a Batch.
        - If it is a dict, replace it with a corresponding Resource instance.
        """
        if not resource_type:
            return preloaded
        elif not many:
            if is_resource(preloaded):
                return preloaded
            elif isinstance(preloaded, dict):
                if 'id' in preloaded:
                    preloaded[ID] = preloaded.pop('id')
                if 'rev' in preloaded:
                    preloaded[REV] = preloaded.pop('rev')
                self.preload_resolver_state(resource_type, preloaded)
                return resource_type(state=preloaded)
            elif (
                (preloaded and isinstance(preloaded, str)) and
                (preloaded[0] == '{' and preloaded[-1] == '}')
            ):
                try:
                    data = self._app.json.decode(preloaded)
                    return resource_type(data)
                except ValueError:
                    return None
            else:
                return resource_type.get(_id=preloaded)
        elif is_sequence(preloaded):
            if isinstance(preloaded, set):
                preloaded = list(preloaded)
            if is_resource(preloaded[0]):
                return resource_type.Batch(resources=preloaded)
            elif isinstance(preloaded[0], dict):
                return resource_type.Batch(
                    resource_type(
                        state=self.preload_resolver_state(resource_type, record)
                    ).merge(
                        _id=None, _rev=None
                    )
                    for record in preloaded
                )
            else:
                return resource_type.get_many(_ids=preloaded)

    def preload_resolver_state(
        self, resource_type: Type['Resource'], preloaded: Dict
    ):
        # remove any items in the preloaded dict if the corresponding field is
        # designated "protected".
        for k in resource_type.ravel.protected_field_names:
            if k in preloaded:
                del preloaded[k]

        for resolver in resource_type.ravel.resolvers.values():
            if resolver.name in resource_type.ravel.resolvers.fields:
                continue
            preloaded_obj = preloaded.get(resolver.name)
            if preloaded_obj:
                target_resource_class = resolver.target
                target_batch_class = target_resource_class.Batch
                if resolver.many:
                    dict_list = preloaded_obj
                    preloaded[resolver.name] = target_batch_class(
                        target_resource_class(
                            state=self.preload_resolver_state(resolver.target, x)
                        ) for x in dict_list
                    )
                else:
                    state_dict = preloaded_obj
                    preloaded[resolver.name] = target_resource_class(
                        state=self.preload_resolver_state(resolver.target, state_dict)
                    )

        return preloaded



import traceback
import inspect

from typing import Dict
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs

from ravel.util.json_encoder import JsonEncoder
from ravel.util.loggers import console

from .abstract_http_server import AbstractHttpServer, Endpoint


class HttpServer(AbstractHttpServer):
    """
    An HTTP server, using the HTTPServer from the standard library. It sends and
    receives JSON data.
    """

    class Handler(BaseHTTPRequestHandler):
        def process(self):
            try:
                result = self.process_request()
                self.write_headers(status_code=200)
                self.encode_response(result)
            except Exception:
                self.write_headers(status_code=400)
                self.encode_response({
                    'error': 'bad request',
                    'traceback': traceback.format_exc().split('\n')
                })

        def write_headers(self, status_code: int):
            self.send_response(status_code)
            self.send_header('Content-type', 'application/json')
            self.end_headers()

        def process_request(self):
            url = urlparse(self.path)
            http_method = self.command
            body = self.read_body_json_object()
            params = self.extract_query_params(url)

            # Create a merged dict of body and params vars
            # for use in the JsonServer on_request method:
            arguments = body.copy()
            arguments.update(params)

            # Apply action callable
            return self.service.route(
                http_method, url.path, args=(arguments, )
            )

        def encode_response(self, result_data):
            if result_data is not None:
                result = {'data': result_data}
                result_json_str = self.service.json_encode(result)
                self.wfile.write(bytes(result_json_str, 'utf8'))

        def read_body_json_object(self):
            body = {}
            content_len = int(self.headers.get('content-length', 0))
            if content_len:
                body_bytes = self.rfile.read(content_len)
                if body_bytes:
                    body = JsonEncoder.decode(body_bytes.decode())
            return body

        def extract_query_params(self, url) -> Dict:
            params = parse_qs(url.query) if url.query else {}
            for k, v in params.items():
                if len(v) == 1:
                    params[k] = v[0]
            return params

        def do_GET(self):
            self.process()

        def do_PUT(self):
            self.process()

        def do_POST(self):
            self.process()

        def do_PATCH(self):
            self.process()

        def do_DELETE(self):
            self.process()

        def do_OPTIONS(self):
            self.process()

    def __init__(self, json_encode=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handler = type('Handler', (self.Handler, ), {'service': self})
        self.json_encode = json_encode or JsonEncoder().encode
        self.server = None

    def on_bootstrap(self, host, port):
        super.on_bootstrap(host, port)
        self.server = HTTPServer((self.host, self.port), self.handler)

    def on_start(self):
        console.info(
            f'HTTP server listening on http://{self.host}:{self.port}'
        )
        self.server.serve_forever()

    def on_request(self, action, arguments: Dict) -> dict:
        args, kwargs = [], {}

        for idx, (k, param) in enumerate(action.signature.parameters.items()):
            if not idx:
                continue
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.default is inspect._empty:
                    args.append(arguments.get(k))
                else:
                    kwargs[k] = arguments.get(k)
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                args.append(arguments.get(k))
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                kwargs[k] = arguments.get(k)
        return (args, kwargs)



import re

import sqlalchemy as sa
import pytz

from sqlalchemy import TypeDecorator
from sqlalchemy.dialects.postgresql import ARRAY


class ArrayOfEnum(TypeDecorator):
    """
    See: https://docs.sqlalchemy.org/en/13/dialects/postgresql.html
    """

    impl = ARRAY
    re_pg_array = re.compile(r'^{(.*)}$')

    def bind_expression(self, bindvalue):
        return sa.cast(bindvalue, self)

    def result_processor(self, dialect, coltype):
        super_rp = super().result_processor(dialect, coltype)

        def processor(value):
            result = []
            if value is not None:
                result = super_rp(self._handle_raw_string(value))
            return result

        return processor

    def _handle_raw_string(self, value):
        match = self.re_pg_array.match(value)
        processed_values = []
        if match:
            inner = match.group(1)
            if inner:
                processed_values = inner.split(',')
        return processed_values


class UtcDateTime(sa.TypeDecorator):

    impl = sa.DateTime

    def process_result_value(self, value, dialect):
        if value:
            return value.replace(tzinfo=pytz.utc)
        return value




import uuid

from typing import Text, Tuple, Callable, Type

from appyratus.schema import Schema, Field, fields

from ravel.util.misc_functions import get_class_name

# TODO: in Resource meta class, recurse through nested schemas for Id fields
#       for resolving them latar

class Id(fields.Field):
    """
    This is a special Field recognized internally by Ravel. When an Application
    bootstraps, all Id fields declared on Resource classes are replaced with a
    concrete Field class.
    """

    def __init__(self, target: Type['Resource'] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.target_resource_type_name = None
        self.target_resource_type_callback = None
        self.target_resource_type = None

        if isinstance(target, type):
            self.target_resource_type = target
        elif isinstance(target, str):
            self.target_resource_type_name = target
        elif callable(target):
            self.target_resource_type_callback = target

    def resolve_target_resource_type(self, app: 'Application') -> Type['Resource']:
        if self.target_resource_type is None:
            if self.target_resource_type_callback is not None:
                app.inject(self.target_resource_type_callback)
                self.target_resource_type = self.target_resource_type_callback()
            else:
                self.target_resource_type = app[self.target_resource_type_name]
        return self.target_resource_type

    # TODO: rename replace_self_in_resource_type
    def replace_self_in_resource_type(
        self,
        app: 'Application',
        source_resource_type: Type['Resource'],
    ):
        # compute the replacement field to use in place of Id
        target_resource_type = self.resolve_target_resource_type(app)
        target_id_field_type = type(target_resource_type._id.resolver.field)
        replacement_field = target_id_field_type(
            name=self.name,
            source=self.source,
            required=self.required,
            nullable=self.nullable,
            meta=self.meta,
        )
        # dynamically replace the existing field in Schema class
        source_resource_type.Schema.replace_field(replacement_field)

        # update existing FieldResolver to use the replacement field
        field_resolver = source_resource_type.ravel.resolvers.fields[self.name]
        field_resolver.field = replacement_field

    def process(self, value):
        raise NotImplementedError()

    def __repr__(self):
        return ('Id()')


# for import convenience:
fields.Id = Id


import inspect

from typing import Text, Set
from collections import defaultdict

from ravel.util.loggers import console
from ravel.util import is_resource, is_batch, get_class_name
from ravel.resolver.resolver import Resolver
from ravel.batch import BatchResolverProperty


class Relationship(Resolver):
    def __init__(self, join, eager=True, order_by=None, *args, **kwargs):
        self._cb_order_by = order_by
        self._order_by = None
        self._eager = eager
        if callable(join):
            self._cb_joins = join
            self._join_sequence = []
        else:
            self._cb_joins = None
            self._join_sequence = join

        super().__init__(*args, **kwargs)

    @classmethod
    def tags(cls) -> Set[Text]:
        return {'relationships'}

    @classmethod
    def priority(cls) -> int:
        return 10

    def dump(self, dumper: 'Dumper', value):
        return dumper.dump(value)

    def on_copy(self, copy):
        copy._cb_order_by = self._cb_order_by
        copy._order_by = None
        copy._eager = self._eager
        copy._cb_joins = self._cb_joins
        copy._join_sequence = self._join_sequence

    def on_bind(self):
        def arg_count(func):
            return len(inspect.signature(func).parameters)

        if self._cb_joins is not None:
            self.app.inject(self._cb_joins)
        if self._cb_order_by is not None:
            self.app.inject(self._cb_order_by)
            if arg_count(self._cb_order_by) > 0:
                self._order_by = self._cb_order_by(self.owner)
            else:
                self._order_by = self._cb_order_by()

        if arg_count(self._cb_joins) > 0:
            pairs = self._cb_joins(self.owner)
        else:
            pairs = self._cb_joins()

        if not isinstance(pairs[0], (list, tuple)):
            pairs = [pairs]

        self._join_sequence = [Join(self, l, r) for l, r in pairs]

        self.target = self._join_sequence[-1].right_loader.owner
        self.many = self._join_sequence[-1].right_many

    def pre_resolve(self, resource, request):
        if self.app.is_simulation:
            # do nothing because, when simulating, we don't need
            # to waste time trying to fetch data.
            return

        source = resource
        final_join = self._join_sequence[-1]
        results = []

        # build sequence of "join queries" to execute, resolving
        # the target Resource type in the final query.
        for j1, j2 in zip(self._join_sequence, self._join_sequence[1:]):
            query = j1.build(source).select(j2.left_field.name)

            if query is None:
                request.result = None
                return

            result = query.execute()
            results.append(result)
            source = result

        # build final query and merge in query parameters
        # passed in through the request
        query = final_join.build(source)
        if query is None:
            request.result = None
            return

        if self._eager:
            query.select(self.target.ravel.resolvers.fields.keys())

        query.merge(request, in_place=True)

        if self._order_by:
            query.order_by(self._order_by)

        result = query.execute(first=not self.many)
        results.append(result)

        # the resolver's returned result is the result from
        # the final query in the "join query" sequence.
        request.result = results[-1]

    def pre_resolve_batch(self, batch, request):
        if self.app.is_simulation:
            # do nothing because, when simulating, we don't need
            # to waste time trying to fetch data.
            return

        mappings = []
        source = batch

        for j1, j2 in zip(
            self._join_sequence,
            self._join_sequence[1:] + [None]
        ):
            query = j1.build(source)

            if query is None:
                request.result = {}
                return

            if j2 is not None:
                # we come here for each Join object except the last
                # in the sequence.
                query.select(j2.left_field.name)
            else:
                # for the final query, merge in query parameters
                # passed in through the request.
                query.merge(request)
                if self._order_by:
                    query.order_by(self._order_by)

            value_2_queried_resource = defaultdict(set)
            queried_resources = query.execute()

            for res in queried_resources:
                right_value = res[j1.right_field.name]
                value_2_queried_resource[right_value].add(res)

            mapping = {}
            for source_res in source:
                source_value = source_res[j1.left_field.name]
                mapping[source_res] = value_2_queried_resource[source_value]

            source = queried_resources
            mappings.append(mapping)

        def extract(key, mappings, index):
            values = mappings[index]
            if index == len(mappings) - 1:
                return values.get(key) or []
            else:
                results = []
                for res in values[key]:
                    extracted_values = extract(res, mappings, index + 1)
                    results.extend(extracted_values)
                return results

        request.result = {}

        for res in batch:
            extracted_resources = extract(res, mappings, 0)
            if self.many:
                request.result[res] = self.target.Batch(extracted_resources)
            else:
                request.result[res] = (
                    list(extracted_resources)[0] if extracted_resources
                    else None
                )

    def on_resolve_batch(self, batch, request):
        return request.result

    def on_simulate(self, resource, request):
        from ravel import Query

        query = Query(request=request)
        if len(self._join_sequence) == 1:
            join = self._join_sequence[0]
            joined_value = getattr(resource, join.left_field.name)
            query.where(join.right_loader_property == joined_value)
        entity = query.execute(first=not self.many)

        return entity


class Join(object):
    def __init__(self, relationship, left, right):
        self.relationship = relationship
        self.left_loader_property = left
        self.left_loader = left.resolver
        self.left_field = left.resolver.field
        self.right_many = False

        if isinstance(right, BatchResolverProperty):
            # in this case, the right-hand field in the join is specified
            # through Batch, like `User.Batch._id`. This is used to indicate
            # that the final query in the Relationship returns "many". This
            # information is already handled before we get here, so at this
            # point, we just replace the batch resolver property with the
            # non-batch one.
            right = getattr(right.resolver.owner, right.resolver.name)
            self.right_many = True

        self.right_loader_property = right
        self.right_loader = right.resolver
        self.right_field = right.resolver.field

    def __repr__(self):
        lhs = (
            f'{get_class_name(self.left_loader.owner)}.'
            f'{self.left_loader.name}'
        )
        rhs = (
            f'{get_class_name(self.right_loader.owner)}.'
            f'{self.right_loader.name}'
        )
        return (
            f'{get_class_name(self)}(from={lhs}, to={rhs})'
        )

    def build(self, source) -> 'Query':
        query = self.right_loader_property.resolver.owner.select()

        if is_resource(source):
            source_value = getattr(source, self.left_field.name)
            if not source_value:
                console.debug(
                    message=(
                        f'relationship {get_class_name(source)}.'
                        f'{self.relationship.name} aborting execution'
                        f'because {self.left_field.name} is None'
                    ),
                    data={
                        'resource': source._id,
                    }
                )
                # NOTE: ^ if you don't want this, then clear the field from
                # source using source.clean(field_name)
                return None
            query.where(self.right_loader_property == source_value)
        else:
            assert is_batch(source)
            left_field_name = self.left_field.name
            source_values = {getattr(res, left_field_name) for res in source}
            if not source_values:
                console.warning(
                    message=(
                        f'relationship {get_class_name(source)}.'
                        f'{self.relationship.name} aborting query '
                        f'because {self.left_field.name} is empty'
                    ),
                    data={
                        'resources': source._id,
                    }
                )
                # NOTE: ^ if you don't want this, then clear the field from
                # source using source.clean(field_name)
                return None
            query.where(self.right_loader_property.including(source_values))

        return query



import inspect
import uuid

from typing import Text, Tuple, List, Set, Dict, Type, Union, Optional
from collections import defaultdict
from datetime import datetime
from copy import deepcopy
from threading import local

from appyratus.utils.dict_utils import DictObject
from appyratus.utils.type_utils import TypeUtils

from ravel.exceptions import NotBootstrapped, ValidationError
from ravel.store import Store, SimulationStore
from ravel.util.dirty import DirtyDict
from ravel.util.loggers import console
from ravel.util import (
    is_batch,
    is_resource,
    is_resource_type,
    is_sequence,
    get_class_name,
)
from ravel.schema import Field, Schema, Id, fields
from ravel.dumper import Dumper, SideLoadedDumper, DumpStyle
from ravel.batch import Batch
from ravel.constants import (
    IS_RESOURCE,
    ID,
    REV,
)

from ravel.query.query import Query
from ravel.query.predicate import PredicateParser
from ravel.resolver.resolver import Resolver
from ravel.resolver.decorators import field as field_decorator
from ravel.resolver.resolver_decorator import ResolverDecorator
from ravel.resolver.resolver_property import ResolverProperty
from ravel.resolver.resolver_manager import ResolverManager
from ravel.resolver.resolvers.loader import Loader
from ravel.entity import Entity


class ResourceMeta(type):
    def __init__(cls, name, bases, dct):
        declared_fields = cls._initialize_class_state()
        cls._inherit_resolvers(bases)
        cls._build_schema_type(declared_fields, bases)
        cls.Batch = Batch.factory(cls)

    def _inherit_resolvers(cls, base_classes):
        for base_class in base_classes:
            if is_resource_type(base_class):
                for key, resolver in base_class.ravel.resolvers.items():
                    copied_resolver = resolver.copy(new_owner=cls)
                    resolver_property = ResolverProperty(copied_resolver)
                    cls.ravel.resolvers.register(copied_resolver)
                    setattr(cls, key, resolver_property)

    def _initialize_class_state(resource_type):
        setattr(resource_type, IS_RESOURCE, True)

        resource_type.ravel = DictObject()
        resource_type.ravel.local = local()
        resource_type.ravel.app = None
        resource_type.ravel.resolvers = ResolverManager()
        resource_type.ravel.foreign_keys = {}
        resource_type.ravel.is_abstract = resource_type._compute_is_abstract()
        resource_type.ravel.is_bootstrapped = False
        resource_type.ravel.is_bound = False
        resource_type.ravel.schema = None
        resource_type.ravel.defaults = {}
        resource_type.ravel.protected_field_names = set()
        resource_type.ravel.predicate_parser = PredicateParser(resource_type)
        resource_type.ravel.virtual_fields = {}

        return resource_type._process_fields()

    def _process_fields(cls):
        fields = {}
        for k, v in inspect.getmembers(cls):
            if isinstance(v, ResolverDecorator):
                resolver_property = v.build_property(owner=cls, name=k)
                resolver = resolver_property.resolver
                cls.ravel.resolvers.register(resolver_property.resolver)
                setattr(cls, k, resolver_property)
                if isinstance(resolver, Loader):
                    fields[k] = resolver.field
            elif isinstance(v, Field):
                field = v
                field.name = k
                fields[k] = field
                resolver_type = field.meta.get('resolver_type') or Loader
                resolver_property = resolver_type.build_property(
                    kwargs=dict(owner=cls, field=field, name=k, target=cls),
                )
                cls.ravel.resolvers.register(resolver_property.resolver)
                setattr(cls, k, resolver_property)
        return fields

    def _compute_is_abstract(resource_type):
        is_abstract = False
        if hasattr(resource_type, '__abstract__'):
            is_abstract = bool(resource_type.__abstract__())
            delattr(resource_type, '__abstract__')
        return is_abstract

    def _build_schema_type(resource_type, fields, base_types):
        fields = fields.copy()
        inherited_fields = {}

        # inherit fields and defaults from base Resource classes
        for base_type in base_types:
            if is_resource_type(base_type):
                inherited_fields.update(base_type.Schema.fields)
                resource_type.ravel.defaults.update(base_type.ravel.defaults)
            else:
                base_fields = resource_type._copy_fields_from_mixin(base_type)
                inherited_fields.update(base_fields)

        fields = dict(inherited_fields, **fields)

        # perform final processing now that we have all direct and
        # inherited fields in one dict.
        default_protected = resource_type.__protected__()
        for k, field in fields.items():
            if field.meta.get('ravel_on_resolve'):
                resource_type.ravel.virtual_fields[k] = field
            if field.meta.get('protected', default_protected) or k in {REV}:
                resource_type.ravel.protected_field_names.add(k)
            if k in inherited_fields:
                resolver_type = field.meta.get('resolver_type') or Loader
                resolver_property = resolver_type.build_property(kwargs=dict(
                    owner=resource_type,
                    field=field,
                    name=k,
                    target=resource_type,
                ))
                resource_type.ravel.resolvers.register(
                    resolver_property.resolver
                )
                setattr(resource_type, k, resolver_property)
            if field.source is None:
                field.source = field.name
            if isinstance(field, Id) and field.name != ID:
                resource_type.ravel.foreign_keys[field.name] = field

        # these are universally required
        assert ID in fields
        assert REV in fields

        fields[ID].nullable = False

        # build new Schema subclass with aggregated fields
        class_name = f'{resource_type.__name__}Schema'
        resource_type.Schema = type(class_name, (Schema, ), fields)

        resource_type.ravel.schema = schema = resource_type.Schema()
        resource_type.ravel.defaults = (
            resource_type._extract_field_defaults(schema)
        )

    def _copy_fields_from_mixin(resource_type, class_obj):
        fields = {}
        is_field = lambda x: isinstance(x, Field)
        for k, field in inspect.getmembers(class_obj, predicate=is_field):
            if k == 'Schema':
                continue
            fields[k] = deepcopy(field)
        return fields

    def _extract_field_defaults(resource_type, schema):

        def make_callable_default(field):
            value = field.default

            def default_callable():
                return deepcopy(value)

            default_callable.__name__ = f'{field.name}_default'
            return default_callable

        defaults = resource_type.ravel.defaults

        for field in schema.fields.values():
            if field.default is not None:
                # move field default into "defaults" dict
                if callable(field.default):
                    defaults[field.name] = field.default
                else:
                    defaults[field.name] = make_callable_default(field)
                # clear it from the schema object once "defaults" dict
                field.default = None
        return defaults


class Resource(Entity, metaclass=ResourceMeta):

    # metaclass attribute stubs:
    ravel = None

    # base resource fields:
    _id = fields.UuidString(default=lambda: uuid.uuid4().hex, nullable=False)
    _rev = fields.String()

    def __init__(self, state=None, **more_state):
        # initialize internal state data dict
        self.internal = DictObject()
        self.internal.state = DirtyDict()
        self.merge(state, **more_state)

        # eagerly generate default ID if none provided
        if ID not in self.internal.state:
            id_func = self.ravel.defaults.get(ID)
            self.internal.state[ID] = id_func() if id_func else None

    def __getitem__(self, key):
        if key in self.ravel.resolvers:
            return getattr(self, key)
        raise KeyError(key)

    def __setitem__(self, key, value):
        if key in self.ravel.resolvers:
            return setattr(self, key, value)
        raise KeyError(key)

    def __delitem__(self, key):
        if key in self.ravel.resolvers:
            delattr(self, key)
        else:
            raise KeyError(key)

    def __iter__(self):
        return iter(self.internal.state)

    def __contains__(self, key):
        return key in self.internal.state

    def __repr__(self):
        name = get_class_name(self)
        dirty = '*' if self.is_dirty else ''
        id_value = self.internal.state.get(ID)
        if id_value is None:
            id_str = '?'
        elif isinstance(id_value, str):
            id_str = id_value[:7]
        elif isinstance(id_value, uuid.UUID):
            id_str = id_value.hex[:7]
        else:
            id_str = repr(id_value)

        return f'{name}({id_str}){dirty}'

    @classmethod
    def __abstract__(cls) -> bool:
        return True

    @classmethod
    def __protected__(cls) -> bool:
        return False

    @classmethod
    def __store__(cls) -> Type[Store]:
        return SimulationStore

    @classmethod
    def on_bootstrap(cls, app, **kwargs):
        pass

    @classmethod
    def on_bind(cls):
        pass

    @classmethod
    def bootstrap(cls, app: 'Application', **kwargs):
        t1 = datetime.now()
        cls.ravel.app = app

        # bootstrap all resolvers owned by this class
        for resolver in cls.ravel.resolvers.values():
            if not resolver.is_bootstrapped():
                resolver.bootstrap(cls.ravel.app)

        # lastly perform custom developer logic
        cls.on_bootstrap(app, **kwargs)
        cls.ravel.local.is_bootstrapped = True

        t2 = datetime.now()
        secs = (t2 - t1).total_seconds()
        console.debug(
            f'bootstrapped {TypeUtils.get_class_name(cls)} '
            f'in {secs:.2f}s'
        )


    @classmethod
    def bind(cls, store: 'Store', **kwargs):
        t1 = datetime.now()
        cls.ravel.local.store = store

        for resolver in cls.ravel.resolvers.values():
            resolver.bind()

        cls.on_bind()
        cls.ravel.is_bound = True

        t2 = datetime.now()
        secs = (t2 - t1).total_seconds()
        console.debug(
            f'bound {TypeUtils.get_class_name(store)} to '
            f'{TypeUtils.get_class_name(cls)} '
            f'in {secs:.2f}s'
        )

    @classmethod
    def is_bootstrapped(cls) -> bool:
        return cls.ravel.local.is_bootstrapped

    @classmethod
    def is_bound(cls) -> bool:
        return cls.ravel.is_bound

    @property
    def app(self) -> 'Application':
        if not self.ravel.app:
            raise NotBootstrapped(
                f'{get_class_name(self)} must be associated '
                f'with a bootstrapped app'
            )
        return self.ravel.app

    @property
    def log(self) -> 'ConsoleLoggerInterface':
        return self.ravel.app.log if self.ravel.app else None

    @property
    def class_name(self) -> Text:
        return get_class_name(self)

    @property
    def store(self) -> 'Store':
        return self.ravel.local.store

    @property
    def is_dirty(self) -> bool:
        return bool(
            self.internal.state.dirty &
            self.ravel.schema.fields.keys()
        )

    @property
    def dirty(self) -> Dict:
        return {
            k: self.internal.state[k]
            for k in self.internal.state.dirty
            if k in self.Schema.fields
        }

    @classmethod
    def generate(
        cls,
        keys: Set[Text] = None,
        values: Dict = None,
        use_defaults=True,
    ) -> 'Resource':
        instance = cls()
        values = values or {}
        keys = keys or set(cls.ravel.schema.fields.keys())
        resolver_objs = Resolver.sort([
            cls.ravel.resolvers[k] for k in keys
            if k not in {REV}
        ])

        instance = cls()

        for resolver in resolver_objs:
            if resolver.name in values:
                value = values[resolver.name]
            elif use_defaults and resolver.name in cls.ravel.defaults:
                func = cls.ravel.defaults[resolver.name]
                value = func()
            else:
                value = resolver.generate(instance)

            instance.internal.state[resolver.name] = value

        return instance

    def set(self, other=None, **values) -> 'Resource':
        return self.merge(other=other, **values)

    def merge(self, other=None, **values) -> 'Resource':
        try:
            if isinstance(other, dict):
                for k, v in other.items():
                    setattr(self, k, v)
            elif isinstance(other, Resource):
                for k, v in other.internal.state.items():
                    setattr(self, k, v)

            if values:
                self.merge(values)

            return self
        except Exception:
            self.app.log.error(
                message=(
                    f'failed to merge object into '
                    f'resource {str(self._id)[:7]}'
                ),
                data={
                    'resource': self._id,
                    'class': self.class_name,
                    'other': other,
                    'values': values,
                }
            )
            raise

    def clean(self, fields=None) -> 'Resource':
        if fields:
            fields = fields if is_sequence(fields) else {fields}
            keys = self._normalize_selectors(fields)
        else:
            keys = set(self.ravel.resolvers.keys())

        if keys:
            self.internal.state.clean(keys=keys)

        return self

    def mark(self, fields=None) -> 'Resource':
        # TODO: rename "mark" method to "touch"
        if fields is not None:
            if not fields:
                return self
            fields = fields if is_sequence(fields) else {fields}
            keys = self._normalize_selectors(fields)
        else:
            keys = set(self.Schema.fields.keys())

        self.internal.state.mark(keys)
        return self

    def dump(
        self,
        resolvers: Optional[Set[Text]] = None,
        style: Optional[DumpStyle] = None
    ) -> Dict:
        """
        Dump the fields of this business object along with its related objects
        (declared as relationships) to a plain ol' dict.
        """
        # get Dumper instance based on DumpStyle (nested, side-loaded, etc)
        dumper = Dumper.for_style(style or DumpStyle.nested)

        if resolvers is not None:
            # only dump resolver state specifically requested
            keys_to_dump = self._normalize_selectors(resolvers)
        else:
            # or else dump all instance state
            keys_to_dump = list(self.internal.state.keys())

        dumped_instance_state = dumper.dump(self, keys=keys_to_dump)
        return dumped_instance_state

    def copy(self) -> 'Resource':
        """
        Create a clone of this Resource
        """
        clone = type(self)(state=deepcopy(self.internal.state))
        return clone.clean()

    def validate(self, resolvers: Set[Text] = None, strict=False) -> Dict:
        """
        Validate an object's loaded state data. If you need to check if some
        state data is loaded or not and raise an exception in case absent,
        use self.require.
        """
        errors = {}
        resolver_names_to_validate = (
            resolvers or set(self.ravel.resolvers.keys())
        )
        for name in resolver_names_to_validate:
            resolver = self.ravel.resolvers[name]
            if name not in self.internal.state:
                console.warning(
                    message=f'skipping {name} validation',
                    data={'reason': 'not loaded'}
                )
            else:
                value = self.internal.state.get(name)
                if value is None and not resolver.nullable:
                    errors[name] = 'not nullable'
                if name in self.ravel.schema.fields:
                    field = self.ravel.schema.fields[name]
                    _value, error = field.process(value)
                    if error is not None:
                        errors[name] = error

        if strict and errors:
            console.error(
                message='validation error',
                data={'errors': errors}
            )
            raise ValidationError('see error log for details')

        return errors

    def require(
        self,
        resolvers: Set[Text] = None,
        use_defaults=True,
        strict=False,
        overwrite=False,
    ) -> Set[Text]:
        """
        Checks if all specified resolvers are present. If they are required
        but not present, an exception will be raised for `strict` mode;
        otherwise, a set of the missing resolver names is returned.
        """
        if isinstance(resolvers, str):
            resolvers = {resolvers}
        required_resolver_names = (
            resolvers or set(
                k for k in self.ravel.resolvers.keys()
                if self.ravel.resolvers[k].required
            )
        )
        missing_resolver_names = set()
        defaults = self.ravel.defaults
        for name in required_resolver_names:
            resolver = self.ravel.resolvers[name]
            if overwrite or name not in self.internal.state:
                if use_defaults and name in defaults:
                    value = defaults[name]()
                    if value is None and not resolver.nullable:
                        raise ValidationError(f'{name} not nullable')
                    self[name] = value
                else:
                    missing_resolver_names.add(name)

        if strict and missing_resolver_names:
            console.error(
                message=f'{self.class_name} missing required data',
                data={'missing': missing_resolver_names}
            )
            raise ValidationError('see error log for details')

        return missing_resolver_names

    def resolve(self, resolvers: Union[Text, Set[Text]] = None) -> 'Resource':
        """
        Execute each of the resolvers, specified by name, storing the results
        in `self.internal.state`.
        """
        if self._id is None:
            return self

        if isinstance(resolvers, str):
            resolvers = {resolvers}
        elif not resolvers:
            resolvers = set(self.ravel.resolvers.keys())

        # execute all requested resolvers
        for k in resolvers:
            resolver = self.ravel.resolvers.get(k)
            if resolver is not None:
                if k in self.ravel.resolvers.fields:
                    # field loader resolvers are treated specially to overcome
                    # the limitation of Resolver.target always expecte to be a
                    # Resource class.
                    # if resolvers == {'session_state'}:
                    #     print(resolvers)
                    #     import ipdb; ipdb.set_trace()
                    obj = resolver.resolve(self)
                    setattr(self, k, getattr(obj, k))
                else:
                    setattr(self, k, resolver.resolve(self))

        # clean the resolved values so they arent't accidently saved on
        # update/create, as we just fetched them from the store.
        self.clean(resolvers)

        return self

    def refresh(self, keys: Union[Text, Set[Text]] = None) -> 'Resource':
        """
        For any field currently loaded in the resource, fetch a fresh copy
        from the store.
        """
        if not keys:
            keys = set(self.internal.state.keys())
        elif isinstance(keys, str):
            keys = {keys}

        return self.resolve(keys)

    def unload(self, keys: Set[Text] = None) -> 'Resource':
        """
        Remove the given keys from field data and/or relationship data.
        """
        if keys:
            if isinstance(keys, str):
                keys = {keys}
                keys = self._normalize_selectors(keys)
        else:
            keys = set(
                self.internal.state.keys() |
                self.ravel.resolvers.keys()
            )
        for k in keys:
            if k in self.internal.state:
                del self.internal.state[k]

        return self

    def is_loaded(self, resolvers: Union[Text, Set[Text]]) -> bool:
        """
        Are all given field and/or relationship values loaded?
        """
        if resolvers:
            if isinstance(resolvers, str):
                resolvers = {resolvers}
                keys = self._normalize_selectors(resolvers)
        else:
            keys = set(
                self.internal.state.keys() |
                self.ravel.resolvers.keys()
            )

        for k in keys:
            is_key_in_data = k in self.internal.state
            is_key_in_resolvers = k in self.ravel.resolvers
            if not (is_key_in_data or is_key_in_resolvers):
                return False

        return True

    def _prepare_record_for_create(self, keys_to_save: Optional[Set[Text]] = None):
        """
        Prepares a a Resource state dict for insertion via DAL.
        """
        # extract only those elements of state data that correspond to
        # Fields declared on this Resource class.
        if ID not in self.internal.state:
            self._id = self.ravel.local.store.create_id(self.internal.state)

        # when inserting or updating, we don't want to write the _rev value on
        # accident. The DAL is solely responsible for modifying this value.
        if REV in self.internal.state:
            del self.internal.state[REV]

        record = {}
        keys_to_save = keys_to_save or set(self.internal.state.keys())
        keys_to_save |= self.ravel.defaults.keys()
        keys_to_save &= self.ravel.resolvers.fields.keys()

        for key in keys_to_save:
            resolver = self.ravel.resolvers[key]
            default = self.ravel.defaults.get(key)
            if key not in self.internal.state:
                if default is not None:
                    self.internal.state[key] = value = default()
                    record[key] = value
                elif resolver.required:
                    raise ValidationError(f'{key} is a required field')
            else:
                value = self.internal.state[key]
                if value is None and (not resolver.nullable):
                    if default:
                        self.internal.state[key] = value = default()
                    # if the value is still none, just remove it from
                    # the state dict instead of raising
                    if self.internal.state[key] is None:
                        console.warning(
                            message=(
                                f'resolved None for {resolver} but not '
                                f'nullable'
                            ),
                            data={
                                'resource': self._id,
                                'class': self.class_name,
                                'field': key,
                            }
                        )
                        del self.internal.state[key]
                        continue
                record[key] = self.internal.state[key]

        return record

    @staticmethod
    def _normalize_selectors(selectors: Set):
        keys = set()
        for k in selectors:
            if isinstance(k, str):
                keys.add(k)
            elif isinstance(k, ResolverProperty):
                keys.add(k.name)
        return keys

    # CRUD Methods

    @classmethod
    def select(
        cls,
        *resolvers: Tuple[Text],
        parent: 'Query' = None,
        request: 'Request' = None,
    ) -> 'Query':
        query = Query(target=cls, request=request, parent=parent)
        query.select(resolvers)
        cls.on_select(query)
        query.callbacks.append(cls.post_select)
        return query

    def create(self, data: Dict = None, **more_data) -> 'Resource':
        self.pre_create()

        data = dict(data or {}, **more_data)
        if data:
            self.merge(data)

        prepared_record = self._prepare_record_for_create()
        prepared_record.pop(REV, None)

        self.on_create(prepared_record)

        created_record = self.ravel.local.store.dispatch(
            'create', (prepared_record, )
        )

        self.merge(created_record)
        self.clean()
        self.post_create()

        return self

    def setdefault(self, key, value):
      if key in self.internal.state:
        return self[key]
      else:
        self[key] = value
        return value

    @classmethod
    def get(cls, _id, select=None) -> Optional[Union['Resource', 'Batch']]:
        if _id is None:
            return None

        if is_sequence(_id):
            return cls.get_many(_ids=_id, select=select)

        if not select:
            select = set(cls.ravel.schema.fields.keys())
        elif not isinstance(select, set):
            select = set(select)

        select |= {ID, REV}
        select -= cls.ravel.virtual_fields.keys()

        cls.on_get(_id, select)

        state = cls.ravel.local.store.dispatch(
            'fetch', (_id, ), {'fields': select}
        )

        resource = cls(state=state).clean() if state else None

        cls.post_get(resource)

        return resource

    @classmethod
    def get_many(
        cls,
        _ids: List = None,
        select=None,
        offset=None,
        limit=None,
        order_by=None,
    ) -> 'Batch':
        """
        Return a list of Resources in the store.
        """
        if not _ids:
            return cls.Batch()

        if not select:
            select = set(cls.ravel.schema.fields)
        elif isinstance(select, set):
            select = set(select)

        select |= {ID, REV}
        select -= cls.ravel.virtual_fields.keys()

        if not (offset or limit or order_by):
            store = cls.ravel.local.store
            args = (_ids, )
            kwargs = {'fields': select}
            states = store.dispatch('fetch_many', args, kwargs).values()
            cls.on_get_many(_ids, select)
            batch = cls.Batch(
                cls(state=state).clean() for state in states
                if state is not None
            )
            cls.post_get_many(batch)
        else:
            query = cls.select(select).where(cls._id.including(_ids))
            query = query.order_by(order_by).offset(offset).limit(limit)
            cls.on_select(query)
            batch = query.execute()
            cls.post_select(query, batch)

        return batch

    @classmethod
    def get_all(
        cls,
        select: Set[Text] = None,
        order_by: List['OrderBy'] = None,
        offset: int = None,
        limit: int = None,
    ) -> 'Batch':
        """
        Return a list of all Resources in the store.
        """
        # build the query
        query = cls.select().where(cls._id != None)

        if select:
            query = query.select(select)
        if order_by:
            query = query.order_by(order_by)
        if limit and limit > 0:
            query = query.limit(limit)
        if offset is not None and offset >= 0:
            query = query.offset(offset)

        cls.on_select(query)

        batch = query.execute()

        cls.post_select(query, batch)

        return batch

    def delete(self) -> 'Resource':
        """
        Call delete on this object's store and therefore mark all fields as
        dirty and delete its _id so that save now triggers Store.create.
        """
        self.ravel.local.store.dispatch('delete', (self._id, ))
        self.mark(self.internal.state.keys())
        self._id = None
        self._rev = None
        return self

    @classmethod
    def delete_many(cls, resources: List['Resource']) -> None:
        # extract ID's of all objects to delete and clear
        # them from the instance objects' state dicts
        resource_ids = []
        for resource in resources:
            resource.mark()
            resource_ids.append(resource._id)
            resource._id = None
            resource._rev = None

        if resource_ids:
            store = cls.ravel.local.store
            cls.on_delete_many(resource_ids)
            store.dispatch('delete_many', args=(resource_ids, ))
            cls.post_delete_many(resource_ids)

    @classmethod
    def delete_all(cls) -> None:
        store = cls.ravel.local.store
        store.dispatch('delete_all')

    @classmethod
    def exists(cls, entity: 'Entity') -> bool:
        """
        Does a simple check if a Resource exists by id.
        """
        store = cls.ravel.local.store

        if not entity:
            return False

        if is_resource(entity):
            args = (entity._id, )
        else:
            id_value, errors = cls._id.resolver.field.process(entity)
            args = (id_value, )
            if errors:
                raise ValueError(str(errors))

        return store.dispatch('exists', args=args)

    @classmethod
    def exists_many(cls, entity: 'Entity') -> bool:
        """
        Does a simple check if a Resource exists by id.
        """
        store = cls.ravel.local.store

        if not entity:
            return False

        if is_batch(entity):
            args = (entity._id, )
        else:
            assert is_sequence(entity)
            id_list = entity
            args = (id_list, )
            for id_value in id_list:
                value, errors = cls._id.resolver.field.process(id_value)
                if errors:
                    raise ValueError(str(errors))

        return store.dispatch('exists_many', args=args)

    def save(self, resolvers: Union[Text, Set[Text]] = None, depth=0) -> 'Resource':
        return self.save_many([self], resolvers=resolvers, depth=depth)[0]

    def update(self, data: Dict = None, **more_data) -> 'Resource':
        data = dict(data or {}, **more_data)
        if data:
            self.merge(data)

        raw_record = self.dirty.copy()
        raw_record.pop(REV, None)
        raw_record.pop(ID, None)

        errors = {}
        prepared_record = {}
        for k, v in raw_record.items():
            field = self.Schema.fields.get(k)
            if field is not None:
                if field.name not in self.ravel.virtual_fields:
                    if v is None and field.nullable:
                        prepared_record[k] = None
                    else:
                        prepared_record[k], error = field.process(v)
                        if error:
                            console.error(
                                f'{self} failed validation for {k}: {v}'
                            )
                            errors[k] = error

        self.on_update(prepared_record)

        if errors:
            raise ValidationError(f'update failed for {self}: {errors}')

        updated_record = self.ravel.local.store.dispatch(
            'update', (self._id, prepared_record)
        )

        if updated_record:
            self.merge(updated_record)

        self.clean(prepared_record.keys() | updated_record.keys())
        self.post_update()

        return self

    @classmethod
    def create_many(
        cls,
        resources: List['Resource'],
        fields: Set[Text] = None
    ) -> 'Batch':
        """
        Call `store.create_method` on input `Resource` list and return them in
        the form of a Batch.
        """
        # normalize resources to a list of Resource objects
        records = []
        prepared_resources = []
        for resource in resources:
            if resource is None:
                continue
            if isinstance(resource, dict):
                # convert raw dict into a proper Resource object
                state_dict = resource
                resource = cls(state=state_dict)

            record = resource._prepare_record_for_create(fields)
            records.append(record)

            resource.internal.state.update(record)
            prepared_resources.append(resource)

        if not records:
            return cls.Batch()

        cls.on_create_many(records)

        store = cls.ravel.local.store
        created_records = store.dispatch('create_many', (records, ))

        for resource, record in zip(prepared_resources, created_records):
            resource.merge(record)
            resource.clean()

        # insert the batch to the store
        batch = cls.Batch(prepared_resources)
        cls.post_create_many(batch)

        return batch

    @classmethod
    def update_many(
        cls,
        resources: List['Resource'],
        fields: Set[Text] = None,
        data: Dict = None,
        **more_data
    ) -> 'Batch':
        """
        Call the Store's update_many method on the list of Resources.
        Multiple Store calls may be made. As a preprocessing step, the input
        resource list is partitioned into groups, according to which subset
        of fields are dirty.

        For example, consider this list of resources,

        ```python
        resources = [
            user1,     # dirty == {'email'}
            user2,     # dirty == {'email', 'name'}
            user3,     # dirty == {'email'}
        ]
        ```

        Calling update on this list will result in two paritions:
        ```python
        assert part1 == {user1, user3}
        assert part2 == {user2}
        ```

        A spearate store call to `update_many` will be made for each partition.
        """
        # common_values are values that should be updated
        # across all objects.
        common_values = dict(data or {}, **more_data)

        # in the procedure below, we partition all incoming Resources
        # into groups, grouped by the set of fields being updated. In this way,
        # we issue an update_many datament for each partition in the DAL.
        partitions = defaultdict(cls.Batch)

        fields_to_update = fields

        for resource in resources:
            if resource is None:
                continue
            if common_values:
                resource.merge(common_values)

            partition_key = tuple(resource.dirty.keys())
            partitions[partition_key].append(resource)

        # id_2_copies used to synchronize updated state across all
        # instances that share the same ID.
        id_2_copies = defaultdict(list)

        cls.on_update_many(cls.Batch(resources))

        for partition_key_tuple, resource_partition in partitions.items():
            records, _ids = [], []

            for resource in resource_partition:
                record = resource.dirty.copy()
                record.pop(REV, None)
                record.pop(ID, None)
                if fields_to_update:
                    record = {
                        k: v for k, v in record.items()
                        if (
                            (k in fields_to_update) and
                            (k not in cls.ravel.virtual_fields)
                        )
                    }
                records.append(record)
                _ids.append(resource._id)

            store = cls.ravel.local.store
            updated_records = store.dispatch('update_many', (_ids, records))

            if not updated_records:
                resource_partition.clean(partition_key_tuple)
                continue

            for resource in resource_partition:
                record = updated_records.get(resource._id)
                if record:
                    resource.merge(record)
                    resource.clean(record.keys())

                    # sync updated state across previously encoutered
                    # instances of this resource (according to ID)
                    if resource._id in id_2_copies:
                        for res_copy in id_2_copies[resource._id]:
                            res_copy.merge(record)
                    id_2_copies[resource._id].append(resource)

        updated_resources = cls.Batch(resources)

        cls.post_update_many(updated_resources)

        return updated_resources

    @classmethod
    def save_many(
        cls,
        resources: List['Resource'],
        resolvers: Union[Text, Set[Text]] = None,
        depth: int = 0
    ) -> 'Batch':
        """
        Essentially a bulk upsert.
        """
        def seems_created(resource):
            return (
                (ID in resource.internal.state) and
                (ID not in resource.internal.state.dirty)
            )

        if resolvers is not None:
            if isinstance(resolvers, str):
                resolvers = {resolvers}
            elif not isinstance(resolvers, set):
                resolvers = set(resolvers)
            fields_to_save = set()
            resolvers_to_save = set()
            for k in resolvers:
                if k in cls.ravel.schema.fields:
                    fields_to_save.add(k)
                else:
                    resolvers_to_save.add(k)
        else:
            fields_to_save = None
            resolvers_to_save = set()

        # partition resources into those that are "uncreated" and those which
        # simply need to be updated.
        to_update = []
        to_create = []
        for resource in resources:
            # TODO: merge duplicates
            if not seems_created(resource):
                to_create.append(resource)
            else:
                to_update.append(resource)

        # perform bulk create and update
        if to_create:
            cls.create_many(to_create, fields=fields_to_save)
        if to_update:
            cls.update_many(to_update, fields=fields_to_save)

        retval = cls.Batch(to_update + to_create)

        if depth < 1:
            # base case. do not recurse on Resolvers
            return retval

        # aggregate and save all Resources referenced by all objects in
        # `resource` via their resolvers.
        class_2_objects = defaultdict(set)
        resolvers = cls.ravel.resolvers.by_tag('fields', invert=True)
        for resolver in resolvers.values():
            if resolver.name not in resolvers_to_save:
                continue
            for resource in resources:
                if resolver.name in resource.internal.state:
                    value = resource.internal.state[resolver.name]
                    resolver.on_save(resolver, resource, value)
                    if value:
                        if is_resource(value):
                            class_2_objects[resolver.owner].add(value)
                        else:
                            assert is_sequence(value)
                            class_2_objects[resolver.owner].update(value)
                    elif value is None and not resolver.nullable:
                        raise ValidationError(
                            f'{get_class_name(cls)}.{resolver.name} '
                            f'is required by save'
                        )

        # recursively call save_many for each type of Resource
        for resource_type, resources in class_2_objects.items():
            resource_type.save_many(resources, depth=depth-1)

        return retval

    @classmethod
    def on_select(cls, query: 'Query'):
        pass

    @classmethod
    def post_select(cls, query: 'Query', result: 'Entity'):
        pass

    @classmethod
    def on_get(cls, _id, fields):
        pass

    @classmethod
    def on_get_many(cls, _ids, fields):
        pass

    @classmethod
    def post_get_many(cls, resources):
        pass

    @classmethod
    def post_get(cls, resource: 'Resource'):
        pass

    @classmethod
    def on_update_many(cls, batch):
        pass

    @classmethod
    def post_update_many(cls, batch):
        pass

    @classmethod
    def on_create_many(cls, batch):
        pass

    @classmethod
    def post_create_many(cls, batch):
        pass

    @classmethod
    def on_delete_many(cls, _ids):
        pass

    @classmethod
    def post_delete_many(cls, _ids):
        pass

    @classmethod
    def post_get(cls, resource: 'Resource'):
        pass

    def on_update(self, values: Dict):
        pass

    def post_update(self):
        pass

    def pre_create(self):
        pass

    def on_create(self, values: Dict):
        pass

    def post_create(self):
        pass

    def on_delete(self):
        pass

    def post_delete(self):
        pass


import inspect

from typing import Dict, Tuple, Set, Type, List

from ravel.util.misc_functions import get_class_name
from ravel.util.loggers import console


def virtual(func):
    """
    This decorator is used to mark methods as "virtual" on the base
    Middleware class. This flag is picked up at runtime, and these methods
    are skipped during ravel requrest processing (via Actions).
    """
    func.is_virtual = True
    return func


class MiddlewareError(Exception):
    def __init__(self, middleware, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.middleware = middleware


class Middleware(object):
    def __init__(self, *args, **kwargs):
        self._is_bootstrapped = False

    def __repr__(self):
        return (
            f'Middleware(name={get_class_name(self)})'
        )

    def bootstrap(self, app: 'Application'):
        self._app = app

        console.debug(
            f'bootstrapping {get_class_name(self)} middleware singleton'
        )

        self.on_bootstrap()
        self._is_bootstrapped = True

    @property
    def is_bootstrapped(self) -> bool:
        return self._is_bootstrapped

    @property
    def app(self) -> 'Application':
        """
        The Application using this middleware.
        """
        return self._app

    @property
    def app_types(self) -> Tuple[Type['Application']]:
        """
        Return a tuple of Application class objects for which this middleware
        applies.
        """
        import ravel

        return (ravel.Application, )

    def on_bootstrap(self):
        """
        Developer-defined custom logic triggered while bootstrapping.
        """

    @virtual
    def pre_request(
        self,
        action: 'Action',
        request: 'Request',
        raw_args: Tuple,
        raw_kwargs: Dict
    ):
        """
        In `pre_request`, `raw_args` and `raw_kwargs` contain the raw args and
        kwargs made available to us by the host Application. For example, in a a
        Ravel application backed by a web framework, the contents of `raw_args`
        could be the request and response objects.
        """

    @virtual
    def on_request(
        self,
        action: 'Action',
        request: 'Request',
        processed_args: Tuple,
        processed_kwargs: Dict
    ):
        """
        At the point `on_request` runs, the host application has transformed the
        `raw_args` and `raw_kwargs` into their processed counterparts, which
        represent the args and kwargs expected by the underyling action
        callable.
        """

    @virtual
    def post_request(
        self,
        action: 'Action',
        request: 'Request',
        result,
    ):
        """
        If no exception was raised in the action callable, we come here.
        """

    @virtual
    def post_bad_request(
        self,
        action: 'Action',
        request: 'Request',
        exc: Exception,
    ):
        """
        If any exception was raised in the action callable, we come here.
        """



import sys

from typing import Text, Tuple, List, Set, Dict, Type, Union, Callable
from collections import defaultdict
from copy import deepcopy
from random import randint
from threading import local

from appyratus.utils.dict_utils import DictObject

from ravel.util.loggers import console
from ravel.util.misc_functions import (
    get_class_name,
    flatten_sequence,
)
from ravel.util import (
    is_resource_type, is_batch_type,
    is_resource, is_batch,
)
from ravel.query.order_by import OrderBy
from ravel.query.request import Request
from ravel.query.predicate import (
    Predicate,
    ConditionalPredicate,
    BooleanPredicate,
    OP_CODE,
)

from .resolver_decorator import ResolverDecorator
from .resolver_property import ResolverProperty

class ResolverMeta(type):
    def __init__(cls, name, bases, dict_):
        cls.ravel = DictObject()
        cls.ravel.local = local()
        cls.ravel.local.is_bootstrapped = False



class Resolver(metaclass=ResolverMeta):

    app = None  # TODO: use cls.ravel.app instead

    def __init__(
        self,
        target=None,
        name=None,
        owner=None,
        decorator=None,
        on_resolve=None,
        on_resolve_batch=None,
        on_select=None,
        on_save=None,
        on_set=None,
        on_get=None,
        on_delete=None,
        on_dump=None,
        private=False,
        nullable=True,
        required=False,
        immutable=False,
        many=False,
    ):
        self.name = name
        self.owner = owner
        self.private = private
        self.nullable = nullable
        self.decorator = decorator
        self.required = required
        self.target_callback = None
        self.target = None
        self.schema = None
        self.many = many
        self.immutable = immutable

        self.on_resolve = on_resolve or self.on_resolve
        self.on_resolve_batch = on_resolve_batch or self.on_resolve_batch
        self.on_select = on_select or self.on_select
        self.on_save = on_save or self.on_save
        self.on_dump = on_dump or self.on_dump
        self.on_get = on_get or self.on_get
        self.on_set = on_set or self.on_set
        self.on_delete = on_delete or self.on_delete

        if is_resource_type(target):
            self.target = target
            self.many = False
        elif is_batch(target):
            self.target = target.owner
            self.many = True
        elif target is not None:
            assert callable(target)
            self.target_callback = target

    def __repr__(self):
        return (
            f'{get_class_name(self)}('
            f'target={get_class_name(self.owner)}.{self.name}'
            f')'
        )

    @classmethod
    def is_bootstrapped(cls):
        return getattr(cls.ravel.local, 'is_bootstrapped', False)

    @classmethod
    def property_type(cls):
        return ResolverProperty

    @classmethod
    def build_property(cls, decorator=None, args=None, kwargs=None):
        args = args or tuple()
        kwargs = kwargs or {}
        resolver = cls(*args, **kwargs)
        property_type = cls.property_type()
        return property_type(resolver, decorator=decorator)

    @classmethod
    def build_decorator(cls, *args, **kwargs):
        class decorator_type(ResolverDecorator):
            def __init__(self, *args, **kwargs):
                super().__init__(cls, *args, **kwargs)

        decorator_type.__name__ = f'{get_class_name(cls)}Decorator'
        return decorator_type

    @classmethod
    def priority(self) -> int:
        """
        Proprity defines the order in which Resolvers execute when executing a
        query, in ascending order.
        """
        return sys.maxsize

    @classmethod
    def tags(cls) -> Set[Text]:
        """
        In development, you can access all Resolvers by tag.  For example, if
        you had a User class with a Resolver called "account" whose class was
        tagged with "my_tag", then you could access this resolver in a
        tag-specific dictionary by doing this:

        ```py
        account_resolver = User.resolvers.my_tag['account']

        # equivalent to doing...
        account_resolver = User.account.resolver
        ```
        """
        return {'untagged'}

    @staticmethod
    def sort(resolvers: List['Resolver']) -> List['Resolver']:
        """
        Sort and return the input resolvers as a new list, orderd by priority
        int, ascending. This reflects the relative order of intended execution,
        from first to last.
        """
        return sorted(resolvers, key=lambda resolver: resolver.priority())

    def copy(self, new_owner: Type['Resource'] = None) -> 'Resolver':
        clone = type(self)(
            target=self.target,
            name=self.name,
            owner=new_owner or self.owner,
            decorator=self.decorator,
            on_resolve=self.on_resolve,
            on_resolve_batch=self.on_resolve_batch,
            on_select=self.on_select,
            on_save=self.on_save,
            on_set=self.on_set,
            on_get=self.on_get,
            on_delete=self.on_delete,
            on_dump=self.on_dump,
            private=self.private,
            nullable=self.nullable,
            required=self.required,
            immutable=self.immutable,
            many=self.many,
        )

        clone.target_callback = self.target_callback
        clone._is_bound = False

        self.on_copy(clone)

        return clone

    @classmethod
    def bootstrap(cls, app: 'Application'):
        console.debug(f'bootstrapping {get_class_name(cls)} class')
        cls.app = app
        cls.on_bootstrap()
        cls.ravel.local.is_bootstrapped = True

    def bind(self):
        if self.target_callback:
            self.app.inject(self.target_callback)
            target = self.target_callback()
            if is_resource_type(target):
                self.target = target
                self.many = False
            elif is_batch_type(target):
                self.target = target.ravel.owner
                self.many = True
            else:
                raise Exception('unrecognized target class')

        if not self.owner.ravel.is_abstract:
            self.on_bind()

        if self.target is not None:
            self.schema = self.target.Schema(name=self.name)

        self._is_bound = True

    def resolve(self, entity: 'Entity', request=None):
        if request is None:
            request = Request(self)

        if is_resource(entity):
            return self.resolve_resource(entity, request)
        else:
            assert is_batch(entity)
            return self.resolve_batch(entity, request)

    def resolve_resource(self, resource: 'Resource', request):
        self.pre_resolve(resource, request)

        if self.app.mode == 'normal':
            result = self.on_resolve(resource, request)
        elif self.app.mode == 'simulation':
            result = self.on_simulate(resource, request)

        processed_result = self.post_resolve(resource, request, result)
        request.result = processed_result
        return processed_result

    def resolve_batch(self, batch: 'Batch', request):
        self.pre_resolve_batch(batch, request)

        if self.app.mode == 'normal':
            result = self.on_resolve_batch(batch, request)
        elif self.app.mode == 'simulation':
            result = {res: self.on_simulate(res, request) for res in batch}

        processed_result = self.post_resolve_batch(batch, request, result)
        return processed_result

    def generate(self, instance=None, request=None):
        if instance is None:
            instance = self.owner.generate()

        if request is None:
            request = Request(self)

        self.pre_resolve(instance, request)
        result = self.on_simulate(instance, request)
        processed_result = self.post_resolve(instance, request, result)

        return processed_result

    def dump(self, dumper: 'Dumper', value):
        return value

    @classmethod
    def on_bootstrap(cls):
        pass

    def on_bind(self):
        return

    def on_copy(self, copy: 'Resolver'):
        pass

    def pre_resolve(self, resource, request):
        return

    def on_resolve(self, resource, request):
        raise NotImplementedError()

    def post_resolve(self, resource, request, result):
        return result

    def on_select(self, request):
        return

    def on_get(self, resource, value):
        return

    def on_set(self, resource, old_value, new_value):
        return

    def on_delete(self, resource, value):
        return

    def on_dump(self, request, value):
        pass

    def on_save(self, resource, value):
        pass

    def on_simulate(self, resource, request):
        query = request.resolver.target.select(request=request)
        if self.many:
            return query.execute(simulate=True)
        else:
            return query.execute(first=True, simulate=True)

    def pre_resolve_batch(self, batch, request):
        return

    def on_resolve_batch(self, batch, request):
        return None

    def post_resolve_batch(self, batch, request, result):
        return result
